# Copyright (C) 2025 pdfarranger contributors
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import cairo
import img2pdf
import pikepdf
import traceback

from math import pi
from gi.repository import Gtk, GObject

from .core import PDFRenderer, _img_to_pdf
from .exporter import _set_meta
from .metadata import merge


class ImageExporter:
    """Export to png, jpg or rasterized pdf (with png or jpg images)"""
    def __init__(self, files, pages, metadata, files_out, config, pdfqueue, exportmode, export_msg):
        self.files = files
        self.model = Gtk.ListStore(GObject.TYPE_PYOBJECT)
        for page in pages:
            page.zoom = config.image_ppi() / 72  # pdf is 72 dpi
            page.resample = -1
            self.model.append([page])
        self.ppi = config.image_ppi()
        self.optimize = config.optimize()
        self.greyscale = config.greyscale()
        self.metadata = metadata
        self.files_out = files_out
        self.pdfqueue = pdfqueue
        self.exportmode = exportmode
        self.export_msg = export_msg
        self.rendering_thread = None
        if exportmode in ['SELECTED_TO_PDF_PNG', 'SELECTED_TO_PDF_JPG']:
            self.pdf_out = pikepdf.Pdf.new()
        self.is_saving = True

    def start(self):
        prange = [0, len(self.model) - 1]
        self.rendering_thread = PDFRenderer(self.model, self.pdfqueue, prange, 1, max_nqueue=3)
        self.rendering_thread.connect('update_thumbnail', self.create_page)
        self.rendering_thread.start()

    def join(self, timeout=None):
        if not self.rendering_thread:
            return
        self.rendering_thread.quit = True
        self.rendering_thread.join(timeout)
        self.is_saving = False

    def is_alive(self):
        return self.is_saving

    def create_page(self, _obj, ref, thumbnail, _zoom, _scale, _is_preview):
        if self.rendering_thread.quit:
            return
        if thumbnail is None:
            # Rendering has ended
            if self.exportmode in ['SELECTED_TO_PDF_PNG', 'SELECTED_TO_PDF_JPG']:
                self.save_pdf()
            self.is_saving = False
            return
        path = ref.get_path()
        page = self.model[path][0]
        ind = Gtk.TreePath.get_indices(path)[0]

        w = thumbnail.get_width()
        h = thumbnail.get_height()
        w1, h1 = (h, w) if page.angle in [90, 270] else (w, h)
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, w1, h1)
        cr = cairo.Context(surface)
        cr.set_source_rgb(1, 1, 1)
        cr.rectangle(0, 0, w1, h1)
        cr.fill()

        if page.angle > 0:
            cr.translate(w1 / 2, h1 / 2)
            cr.rotate(page.angle * pi / 180)
            cr.translate(-w / 2, -h / 2)
        cr.set_source_surface(thumbnail)
        cr.paint()

        imgpil = self.surface_to_pil(surface)
        if self.greyscale:
            imgpil = imgpil.convert('L')

        ext = 'png' if self.exportmode in ['SELECTED_TO_PNG', 'SELECTED_TO_PDF_PNG'] else 'jpeg'
        if self.exportmode in ['SELECTED_TO_PNG', 'SELECTED_TO_JPG']:
            self.save_image(imgpil, ext, self.files_out[ind])
        else:
            self.add_to_pdf(imgpil, ext, page.size_in_points())
        self.rendering_thread.nqueue -= 1

    @staticmethod
    def surface_to_pil(surface):
        size = (surface.get_width(), surface.get_height())
        stride = surface.get_stride()
        with surface.get_data() as memory:
            return img2pdf.Image.frombuffer('RGB', size, memory.tobytes(), 'raw', 'BGRX', stride)

    def save_image(self, imgpil, ext, filename):
        try:
            imgpil.save(filename, ext, dpi=(self.ppi, self.ppi), optimize=self.optimize)
        except OSError as e:
            self.exception_handler(e)

    def add_to_pdf(self, imgpil, ext, page_size):
        imgio = img2pdf.BytesIO()
        try:
            imgpil.save(imgio, ext, dpi=(self.ppi, self.ppi), optimize=self.optimize)
        except OSError as e:
            self.exception_handler(e)
            return
        imgio.seek(0)
        pdf = _img_to_pdf([imgio], tmp_dir=None, page_size=page_size)
        src = pikepdf.Pdf.open(pdf)
        self.pdf_out.pages.extend(src.pages)

    def save_pdf(self):
        m = merge(self.metadata, self.files)
        _set_meta(m, [], self.pdf_out)
        try:
            self.pdf_out.save(self.files_out[0])
        except OSError as e:
            self.exception_handler(e)

    def exception_handler(self, e):
        print(traceback.format_exc())
        self.export_msg.put([e, Gtk.MessageType.ERROR])
        self.join()



# Copyright (C) 2008-2017 Konstantinos Poulios, 2018-2019 Jerome Robert
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import pikepdf
import os
import traceback
import sys
import warnings
import tempfile
import io
import gi
import locale
from typing import Any, Dict, List

from . import metadata
from gi.repository import Gtk
gi.require_version("Poppler", "0.18")
from gi.repository import Poppler
import gettext
_ = gettext.gettext

from .core import Page, Sides

# pikepdf.Page.add_overlay()/add_underlay() can't place a page exactly
# if for example LC_NUMERIC=fi_FI
try:
    locale.setlocale(locale.LC_NUMERIC, 'C')
except locale.Error:
    pass  # Gtk already prints a warning

if os.name == 'nt':
    # Work around https://github.com/pdfarranger/pdfarranger/issues/1110
    locale.setlocale(locale.LC_COLLATE, 'C')


def get_blank_doc(pageadder, pdfqueue, tmpdir, size, npages=1):
    """Search pdfqueue for a matching pdf with blank pages. Create it if it does not exist.

    Some notes:
    Blank pdf documents are created prior to export (vs created as needed at export)
    because it will keep code simpler. For example rendering of thumbnails require no
    extra for rendering of a blank page.

    A document with several blank pages is needed if the page number under thumbnail
    need to be something else than 1.
    """
    for i, pdfdoc in enumerate(pdfqueue):
        if size == pdfdoc.blank_size and npages == pdfdoc.document.get_n_pages():
            filename = pdfdoc.copyname
            nfile = i + 1
            return filename, nfile
    filename = _create_blank_page(tmpdir, size, npages)
    doc_data = pageadder.get_pdfdoc(filename, description=None, blank_size=size)
    if doc_data is None:
        return None, None
    nfile = doc_data[1]
    return filename, nfile


def _create_blank_page(tmpdir, size, npages=1):
    """
    Create a temporary PDF file with npages empty pages.
    The size is in PDF unit (1/72 of inch).
    """
    f, filename = make_tmp_file(tmpdir)
    f.add_blank_page(page_size=size)
    for __ in range(npages - 1):
        f.pages.append(f.pages[0])
    f.save(filename)
    return filename


def make_tmp_file(tmpdir):
    fd, filename = tempfile.mkstemp(suffix=".pdf", dir=tmpdir)
    os.close(fd)
    f = pikepdf.Pdf.new()
    return f, filename


def _normalize_rectangle(rect):
    """
    PDF Specification 1.7, 7.9.5, although rectangles are conventionally
    specified by their lower-left and upper-right corners, it is acceptable to
    specify any two diagonally opposite corners. Applications that process PDF
    should be prepared to normalize such rectangles in situations where
    specific corners are required.
    """
    rect = [float(x) for x in rect]
    if rect[0] > rect[2]:
        rect[0], rect[2] = rect[2], rect[0]
    if rect[1] > rect[3]:
        rect[1], rect[3] = rect[3], rect[1]
    return rect


def _intersect_rectangle(rect1, rect2):
    return [
        max(rect1[0], rect2[0]),
        max(rect1[1], rect2[1]),
        min(rect1[2], rect2[2]),
        min(rect1[3], rect2[3]),
    ]


def _mediabox(page, crop=Sides()):
    """ Return the media box for a given page. """
    # PDF files which do not have mediabox default to Portrait Letter / ANSI A
    cmb = page.MediaBox if "/MediaBox" in page else [0, 0, 612, 792]
    cmb = _normalize_rectangle(cmb)
    if "/CropBox" in page:
        # PDF specification §14.11.2.1, "If they do, they are effectively
        # reduced to their intersection with the media box"
        cmb = _intersect_rectangle(cmb, _normalize_rectangle(page.CropBox))

    if crop == Sides():
        return cmb
    angle = page.Rotate if '/Rotate' in page else 0
    rotate_times = int(round(((angle) % 360) / 90) % 4)
    crop_init = crop
    if rotate_times != 0:
        perm = [0, 2, 1, 3]
        for _ in range(rotate_times):
            perm.append(perm.pop(0))
        perm.insert(1, perm.pop(2))
        crop = Sides(*(crop_init[perm[side]] for side in range(4)))
    x1, y1, x2, y2 = [float(x) for x in cmb]
    x1_new = x1 + (x2 - x1) * crop.left
    x2_new = x2 - (x2 - x1) * crop.right
    y1_new = y1 + (y2 - y1) * crop.bottom
    y2_new = y2 - (y2 - y1) * crop.top
    return [x1_new, y1_new, x2_new, y2_new]


def _set_meta(mdata, pdf_input, pdf_output):
    ppae = metadata.PRODUCER not in mdata
    with pdf_output.open_metadata(set_pikepdf_as_editor=ppae) as outmeta:
        if len(pdf_input) > 0:
            metadata.load_from_docinfo(outmeta, pdf_input[0])
        for k, v in mdata.items():
            outmeta[k] = v


def _scale(doc, page, factor):
    """ Scale a page """
    if factor == 1:
        return page
    rotate = 0
    if "/Rotate" in page:
        # We'll set the rotate attribute on the resulting page so we must
        # unset it on the input page before
        rotate = page.Rotate
        page.Rotate = 0
    page_id = len(doc.pages)
    newmediabox = [factor * float(x) for x in page.MediaBox]
    content = "q {} 0 0 {} 0 0 cm /p{} Do Q".format(factor, factor, page_id)
    xobject = pikepdf.Page(page).as_form_xobject()
    new_page = pikepdf.Dictionary(
        Type=pikepdf.Name.Page,
        MediaBox=newmediabox,
        Contents=doc.make_stream(content.encode()),
        Resources={'/XObject': {'/p{}'.format(page_id): xobject}},
        Rotate=rotate,
    )
    # This was needed for pikepdf <= 2.6.0. See https://github.com/pikepdf/pikepdf/issues/174
    # It's also needed with pikepdf 4.2 else we get:
    # RuntimeError: QPDFPageObjectHelper::getFormXObjectForPage called with a direct object
    # when calling as_form_xobject in generate_booklet
    # It's also needed for pikepdf 8.8 but not for pikepdf 6.0.1
    new_page = doc.make_indirect(new_page)
    return new_page


def _update_angle(model_page, source_page, output_page):
    angle = model_page.angle
    angle0 = source_page.Rotate if '/Rotate' in source_page else 0
    if angle != 0:
        new_angle = angle + angle0
        if new_angle >= 360:
            new_angle -= 360
        output_page.Rotate = new_angle


def _apply_geom_transform(pdf_output, new_page, row):
    _update_angle(row, new_page, new_page)
    new_page.MediaBox = _mediabox(new_page, row.crop)
    # add_overlay() & add_underlay() will use TrimBox or CropBox if they exist
    if '/TrimBox' in new_page:
        del new_page.TrimBox
    if '/CropBox' in new_page:
        del new_page.CropBox
    return pikepdf.Page(_scale(pdf_output, new_page, row.scale))


def _apply_geom_transform_job(pdf_output:pikepdf.Pdf, new_page:pikepdf.Page, page:Page) -> None:
    new_page.rotate(page.angle, relative=True)
    new_page.MediaBox = _mediabox(new_page, page.crop)
    # add_overlay() & add_underlay() will use TrimBox or CropBox if they exist
    if '/TrimBox' in new_page:
        del new_page.TrimBox
    if '/CropBox' in new_page:
        del new_page.CropBox
    if page.scale != 1:
        pdf_output.pages.append(new_page)
        new_page.obj.emplace(_scale(pdf_output, pdf_output.pages[-1], page.scale))
        del(pdf_output.pages[-1])



def _remove_unreferenced_resources(pdfdoc):
    try:
        pdfdoc.remove_unreferenced_resources()
    except RuntimeError:
	# Catch "RuntimeError: operation for dictionary attempted on object of
	# type null" with old version PikePDF (observed with 1.17 and 1.19).
	# Blindly catch all RuntimeError is dangerous as this may catch
	# unwanted exception so we print it.
        print(traceback.format_exc())

def warn_dialog(func):
    """ Decorator which redirect warnings and error messages to a gtk MessageDialog """
    class ShowWarning:
        def __init__(self):
            self.buffer = ""

        def __call__(self, message, category, filename, lineno, f=None, line=None):
            s = warnings.formatwarning(message, category, filename, lineno, line)
            if sys.stderr is not None:
                sys.stderr.write(s + '\n')
            self.buffer += str(message) + '\n'

    def wrapper(*args, **kwargs):
        export_msg = kwargs["export_msg"]
        backup_showwarning = warnings.showwarning
        warnings.showwarning = ShowWarning()
        try:
            func(*args, **kwargs)
            if len(warnings.showwarning.buffer) > 0:
                export_msg.put([warnings.showwarning.buffer, Gtk.MessageType.WARNING])
        except Exception as e:
            traceback.print_exc()
            export_msg.put([e, Gtk.MessageType.ERROR])
        finally:
            warnings.showwarning = backup_showwarning

    return wrapper

def export_process(*args, **kwargs):
    """Export PDF in a separate process."""
    warn_dialog(export)(*args, **kwargs)


def _copy_n_transform(pdf_input, pdf_output, pages, quit_flag=None):
    # all pages must be copied to pdf_output BEFORE applying geometrical
    # transformation. See https://github.com/pikepdf/pikepdf/issues/271
    copied_pages = {}
    mediaboxes = []
    # Copy pages from the input PDF files to the output PDF file
    for row in pages:
        if quit_flag is not None and quit_flag.is_set():
            return
        current_page = pdf_input[row.nfile - 1].pages[row.npage - 1]
        mediaboxes.append(_mediabox(current_page))
        _append_page(current_page, copied_pages, pdf_output, row)
        # Layer pages are temporary added after the page they belong to
        for lprow in row.layerpages:
            layer_page = pdf_input[lprow.nfile - 1].pages[lprow.npage - 1]
            _append_page(layer_page, copied_pages, pdf_output, lprow)

    # Apply geometrical transformations in the output PDF file
    i = 0
    for row in pages:
        if quit_flag is not None and quit_flag.is_set():
            return

        pdf_output.pages[i] = _apply_geom_transform(pdf_output, pdf_output.pages[i], row)
        for lprow in row.layerpages:
            i += 1
            pdf_output.pages[i] = _apply_geom_transform(pdf_output, pdf_output.pages[i], lprow)
        i += 1

    # Add overlays and underlays
    for i, row in enumerate(pages):
        # The dest page coordinates and size before geometrical transformations
        dx1, dy1, dx2, dy2 = mediaboxes[i]
        dw, dh = dx2 - dx1, dy2 - dy1

        dpage = pdf_output.pages[i]
        dangle0 = dpage.Rotate if '/Rotate' in dpage else 0
        rotate_times = int(round(((dangle0) % 360) / 90) % 4)
        for lprow in row.layerpages:
            # Rotate the offsets so they are relative to dest page
            offset = lprow.offset.rotated(rotate_times)
            offs_left, offs_right, offs_top, offs_bottom = offset
            x1 = row.scale * (dx1 + dw * offs_left)
            y1 = row.scale * (dy1 + dh * offs_bottom)
            x2 = row.scale * (dx1 + dw * (1 - offs_right))
            y2 = row.scale * (dy1 + dh * (1 - offs_top))
            rect = pikepdf.Rectangle(x1, y1, x2, y2)

            layer_page = pdf_output.pages[i + 1]
            if lprow.laypos == 'OVERLAY':
                pdf_output.pages[i].add_overlay(layer_page, rect)
            else:
                pdf_output.pages[i].add_underlay(layer_page, rect)
            # Remove the temporary added page
            del pdf_output.pages[i + 1]


def _append_page(current_page, copied_pages, pdf_output, row):
    """Add a page to the output pdf. A page that already exist is duplicated."""
    new_page = copied_pages.get((row.nfile, row.npage))
    if new_page is None:
        new_page = current_page
    # let pdf_output adopt new_page
    pdf_output.pages.append(new_page)
    new_page = pdf_output.pages[-1]
    copied_pages[(row.nfile, row.npage)] = new_page
    # Ensure annotations are copied rather than referenced
    # https://github.com/pdfarranger/pdfarranger/issues/437
    if pikepdf.Name.Annots in current_page:
        pdf_temp = pikepdf.Pdf.new()
        pdf_temp.pages.append(current_page)
        indirect_annots = pdf_temp.make_indirect(pdf_temp.pages[0].Annots)
        new_page.Annots = pdf_output.copy_foreign(indirect_annots)


def _transform_job(pdf_output: pikepdf.Pdf, pages: List[Page], quit_flag = None) -> None:
    """ Same as _copy_n_transform, except it doesn't copy. Requires pikepdf >= 8.0 """
    # Fix missing MediaBoxes
    for page in pdf_output.pages:
        if page.mediabox is None:
            page.mediabox = pikepdf.Array((0, 0, 612, 792))

    # We don't need to call _append_page as the Job interface copies pages / annotations as necessary.
    mediaboxes:List[pikepdf.Rectangle] = []
    i = 0
    for page in pages:
        if quit_flag is not None and quit_flag.is_set():
            return
        mediaboxes.append(pikepdf.Rectangle(pdf_output.pages[i].mediabox))
        _apply_geom_transform_job(pdf_output, pdf_output.pages[i], page)
        for lpage in page.layerpages:
            i += 1
            _apply_geom_transform_job(pdf_output, pdf_output.pages[i], lpage)
        i += 1

    # # Add overlays and underlays
    for i, page in enumerate(pages):
        # The dest page coordinates and size before geometrical transformations
        mb = mediaboxes[i]

        # Call to rotate in _apply_geom_transform_job ensures /Rotate exists
        rotate_times = int(round((pdf_output.pages[i].Rotate % 360) / 90) % 4)
        for lpage in page.layerpages:
            # Rotate the offsets so they are relative to dest page
            offset = lpage.offset.rotated(rotate_times)
            offs_left, offs_right, offs_top, offs_bottom = offset
            x1 = page.scale * (mb.llx + mb.width * offs_left)
            y1 = page.scale * (mb.lly + mb.height * offs_bottom)
            x2 = page.scale * (mb.llx + mb.width * (1 - offs_right))
            y2 = page.scale * (mb.lly + mb.height * (1 - offs_top))
            rect = pikepdf.Rectangle(x1, y1, x2, y2)

            if lpage.laypos == 'OVERLAY':
                pdf_output.pages[i].add_overlay(pdf_output.pages[i + 1], rect)
            else:
                pdf_output.pages[i].add_underlay(pdf_output.pages[i + 1], rect)
            # Remove the temporary added page
            del pdf_output.pages[i + 1]


def get_max_pdf_version(pdf_list: List[pikepdf.Pdf]) -> str:
    """Return the highest pdf version used in pdf list"""
    versions = []
    for pdf in pdf_list:
        if pdf is None:
            # Can be None when printing
            continue
        versions.append(pdf.pdf_version)
    return max(*versions)


def export_doc(pdf_input, pages, mdata, files_out, quit_flag, test_mode=False):
    """Same as export() but with pikepdf.PDF objects instead of files"""
    pdf_output = pikepdf.Pdf.new()
    max_version = get_max_pdf_version([pdf_output, *pdf_input])
    _copy_n_transform(pdf_input, pdf_output, pages, quit_flag)
    if quit_flag is not None and quit_flag.is_set():
        return
    if isinstance(files_out[0], str):
        # Only needed when saving to file, not when printing
        mdata = metadata.merge_doc(mdata, pdf_input)
    if len(files_out) > 1:
        for n, page in enumerate(pdf_output.pages):
            if quit_flag is not None and quit_flag.is_set():
                return
            outpdf = pikepdf.Pdf.new()
            _set_meta(mdata, pdf_input, outpdf)
            # works without make_indirect as already applied to this page
            outpdf.pages.append(page)
            _remove_unreferenced_resources(outpdf)
            outpdf.save(files_out[n], min_version=max_version)
    else:
        if isinstance(files_out[0], str):
            if not test_mode:
                _set_meta(mdata, pdf_input, pdf_output)
            _remove_unreferenced_resources(pdf_output)
        if test_mode:
            pdf_output.save(files_out[0], qdf=True, static_id=True, compress_streams=False,
                stream_decode_level=pikepdf.StreamDecodeLevel.all,
                min_version=max_version,
            )
        else:
            pdf_output.save(files_out[0], min_version=max_version)


def _add_json_entries(json: Dict[str, Any], files: List[List[str]], page: Page) -> None:
    """Create an entry for the job json "pages" list."""
    pages_entry = {"file": files[page.nfile - 1][0],  # copyname
                   "range": str(page.npage)}
    if len(files[page.nfile - 1][1]) > 0:
        pages_entry["password"] = files[page.nfile - 1][1]
    json["pages"].append(pages_entry)


def _create_job(files: List[List[str]], pages: List[Page], files_out: List[str], quit_flag=None,
                test_mode: bool = False):
    """ Same as _copy_n_transform, except it use the pikepdf Job interface. Requires pikepdf >= 8.0 """
    # Generate the output PDF file including temporary overlay/ underlay pages. We don't need to call
    # _append_page as the Job interface copies pages / annotations as necessary. We can also delay getting
    # our MediaBoxes until the transformation stage.
    json = dict(outputFile=files_out[0], pages=[], removeUnreferencedResources="yes")
    if test_mode:
        json.update(qdf="", staticId="", compressStreams="n", decodeLevel="all")
    if len(files) > 0 and len(files[0][0]) > 0:
        json["inputFile"] = files[0][0]  # We are treating files [0] as the main document
        if len(files[0][1]) > 0:
            json["password"] = files[0][1]
    else:
        json["inputFile"] = "."

    for page in pages:
        if quit_flag is not None and quit_flag.is_set():
            return None
        _add_json_entries(json, files, page)
        for lpage in page.layerpages:
            # Layer pages are temporarily added after the page they belong to
            _add_json_entries(json, files, lpage)
    return pikepdf.Job(json)


def export_doc_job(pdf_input: List[pikepdf.Pdf], files: List[List[str]], pages: List[Page], mdata, files_out: List[str],
                   quit_flag, test_mode: bool = False) -> None:
    """  Same as export() but uses the pikepdf Job interface. Requires pikedf >= 8.0. """
    job = _create_job(files, pages, files_out, quit_flag, test_mode)
    pdf_output = job.create_pdf()
    max_version = get_max_pdf_version([pdf_output, *pdf_input])

    _transform_job(pdf_output, pages, quit_flag)

    if quit_flag is not None and quit_flag.is_set():
        return
    if isinstance(files_out[0], str):
        # Only needed when saving to file, not when printing
        mdata = metadata.merge_doc(mdata, pdf_input)
    if len(files_out) > 1:
        for n, page in enumerate(pdf_output.pages):
            if quit_flag is not None and quit_flag.is_set():
                return
            outpdf = pikepdf.Pdf.new()
            _set_meta(mdata, pdf_input, outpdf)
            outpdf.pages.append(page)
            _remove_unreferenced_resources(outpdf)
            outpdf.save(files_out[n], min_version=max_version)
    else:
        if isinstance(files_out[0], str) and not test_mode:
            _set_meta(mdata, [pdf_output], pdf_output)
        job.write_pdf(pdf_output)


def export(files, pages, mdata, files_out, config, quit_flag, test_mode=False, **kwargs):
    pdf_input = [
        pikepdf.open(copyname, password=password) for copyname, password in files
    ]
    if config.start_with_empty():
        export_doc(pdf_input, pages, mdata, files_out, quit_flag, test_mode)
    else:
        export_doc_job(pdf_input, files, pages, mdata, files_out, quit_flag, test_mode)


def num_pages(filepath):
    """Get number of pages for filepath."""
    try:
        pdf = pikepdf.Pdf.open(filepath)
    except pikepdf._qpdf.PdfError:
        return None
    npages = len(pdf.pages)
    pdf.close()
    return npages


def generate_booklet(pdfqueue, tmp_dir, pages):
    file, filename = make_tmp_file(tmp_dir)
    content_dict = pikepdf.Dictionary({})
    file_indexes = set()
    for p in pages:
        file_indexes.add(p.nfile)
        for lp in p.layerpages:
            file_indexes.add(lp.nfile)
    source_files = {n-1: pikepdf.open(pdfqueue[n - 1].copyname) for n in file_indexes}
    max_version = max(pdf.pdf_version for pdf in source_files.values())
    _copy_n_transform(source_files, file, pages)
    to_remove = len(file.pages)
    npages = len(pages)
    for i in range(npages//2):
        even = i % 2 == 0
        first_id = -i - 1 if even else i
        second_id = i if even else -i - 1
        if first_id < 0:
            first_id += npages
        if second_id < 0:
            second_id += npages
        first = pages[first_id]
        second = pages[second_id]
        first_foreign = file.pages[first_id]
        second_foreign = file.pages[second_id]
        second_page_size = second.size_in_points()
        first_page_size = first.size_in_points()
        page_size = [max(second_page_size[0], first_page_size[0]) * 2,
                     max(second_page_size[1], first_page_size[1])]

        content_dict[f'/Page{i*2}'] = pikepdf.Page(first_foreign).as_form_xobject()
        content_dict[f'/Page{i*2 + 1}'] = pikepdf.Page(second_foreign).as_form_xobject()
        # See PDF reference section 4.2.3 Transformation Matrices
        tx1 = -first_foreign.MediaBox[0]
        ty1 = -first_foreign.MediaBox[1]
        tx2 = first_page_size[0] - float(second_foreign.MediaBox[0])
        ty2 = -second_foreign.MediaBox[1]
        content_txt = (
            f"q 1 0 0 1 {tx1} {ty1} cm /Page{i*2} Do Q "
            f"q 1 0 0 1 {tx2} {ty2} cm /Page{i*2 + 1} Do Q "
        )

        newpage = pikepdf.Dictionary(
                Type=pikepdf.Name.Page,
                MediaBox=[0, 0, *page_size],
                Resources=pikepdf.Dictionary(XObject=content_dict),
                Contents=pikepdf.Stream(file, content_txt.encode())
            )

        file.pages.append(pikepdf.Page(newpage))
    for __ in range(to_remove):
        del file.pages[0]
    file.save(filename, min_version=max_version)
    return filename


class PrintSettingsWidget(Gtk.Grid):
    def __init__(self, scale_mode, auto_rotate):
        super().__init__(margin=0, row_spacing=6, column_spacing=12, border_width=12)
        lbl = Gtk.Label(_("Scale mode:"), margin=0)
        self.combo = Gtk.ComboBoxText(margin=0)
        self.combo.append("NONE", _("None"))
        self.combo.append("PRINTABLE", _("Fit to Printable Area"))
        self.combo.append("FULL", _("Fit to Full Page"))
        self.combo.set_active_id(scale_mode)
        self.cb = Gtk.CheckButton(label=_("Auto Rotate"), margin=0)
        self.cb.set_active(auto_rotate)
        self.attach(lbl, 0, 1, 1, 1)
        self.attach(self.combo, 1, 1, 2, 1)
        self.attach(self.cb, 0, 2, 2, 1)
        self.show_all()

    def get_scale_mode(self):
        return self.combo.get_active_id()

    def get_auto_rotate(self):
        return self.cb.get_active()


# Adapted from https://stackoverflow.com/questions/28325525/python-gtk-printoperation-print-a-pdf
class PrintOperation(Gtk.PrintOperation):
    MESSAGE=_("Printing…")
    def __init__(self, app):
        super().__init__(embed_page_setup=True, support_selection=True)
        self.app = app
        self.connect("create-custom-widget", self.create_custom_widget)
        self.connect("custom-widget-apply", self.custom_widget_apply)
        self.connect("request-page-setup", self.request_page_setup)
        self.connect("begin-print", self.begin_print, None)
        self.connect("end-print", self.end_print, None)
        self.connect("draw-page", self.draw_page, None)
        self.connect("preview", self.preview, None)
        self.message = self.MESSAGE
        self.scale_mode = self.app.config.scale_mode()
        self.auto_rotate = self.app.config.auto_rotate()
        self.set_use_full_page(self.scale_mode != 'PRINTABLE')
        self.snums = [p.get_indices() for p in reversed(app.iconview.get_selected_items())]
        self.set_has_selection(len(self.snums) > 0)

    def create_custom_widget(self, operation):
        self.set_custom_tab_label(_("Page Handling"))
        return PrintSettingsWidget(self.scale_mode, self.auto_rotate)

    def custom_widget_apply(self, operation, widget):
        self.scale_mode = widget.get_scale_mode()
        self.auto_rotate = widget.get_auto_rotate()
        self.set_use_full_page(self.scale_mode != 'PRINTABLE')

    def request_page_setup(self, operation, print_ctx, page_num, setup):
        if self.auto_rotate:
            w_page = self.pages[page_num].width_in_points()
            h_page = self.pages[page_num].height_in_points()
            if w_page >= h_page:
                setup.set_orientation(Gtk.PageOrientation.LANDSCAPE)
            else:
                setup.set_orientation(Gtk.PageOrientation.PORTRAIT)

    def preview(self, operation, preview_op, print_ctx, parent, user_data):
        self.message = _("Rendering Preview…")

    def begin_print(self, operation, print_ctx, print_data):
        self.app.set_export_state(True, self.message)
        psel = self.get_print_settings().get_print_pages() == Gtk.PrintPages.SELECTION
        nums = self.snums if psel else range(len(self.app.model))
        self.pages = [self.app.model[n][0].duplicate(incl_thumbnail=False) for n in nums]
        self.app.apply_hide_margins_on_pages(self.pages)
        self.set_n_pages(len(self.pages))

        # Create a temporary pdf of the pages to print
        nfiles = set()
        for p in self.pages:
            nfiles.add(p.nfile)
            for lp in p.layerpages:
                nfiles.add(lp.nfile)
        pdf_input = [None] * len(self.app.pdfqueue)
        for nfile in nfiles:
            pdf = self.app.pdfqueue[nfile - 1]
            pdf_input[nfile - 1] = pikepdf.open(pdf.copyname, password=pdf.password)
        self.buf = io.BytesIO()
        export_doc(pdf_input, self.pages, {}, [self.buf], None)
        self.temp_doc = Poppler.Document.new_from_data(self.buf.getvalue())

    def end_print(self, operation, print_ctx, print_data):
        self.app.set_export_state(False)
        self.message = self.MESSAGE
        self.app.config.set_scale_mode(self.scale_mode)
        self.app.config.set_auto_rotate(self.auto_rotate)

    def draw_page(self, operation, print_ctx, page_num, print_data):
        cairo_ctx = print_ctx.get_cairo_context()
        # Poppler context is always 72 dpi
        cairo_ctx.scale(print_ctx.get_dpi_x() / 72, print_ctx.get_dpi_y() / 72)
        if page_num >= len(self.app.model):
            return
        setup = print_ctx.get_page_setup()
        if self.scale_mode == 'PRINTABLE':
            w_paper = setup.get_page_width(Gtk.Unit.POINTS)
            h_paper = setup.get_page_height(Gtk.Unit.POINTS)
        else:
            w_paper = setup.get_paper_width(Gtk.Unit.POINTS)
            h_paper = setup.get_paper_height(Gtk.Unit.POINTS)
        w_page = self.pages[page_num].width_in_points()
        h_page = self.pages[page_num].height_in_points()
        print_scale = self.get_print_settings().get_scale() / 100
        scale = 1
        if self.scale_mode != 'NONE':
            w_scale = w_paper / (w_page * print_scale)
            h_scale = h_paper / (h_page * print_scale)
            scale = min(w_scale, h_scale)
            cairo_ctx.scale(scale, scale)

        # Center page on paper
        dx = w_paper / (scale * print_scale) - w_page
        dy = h_paper / (scale * print_scale) - h_page
        cairo_ctx.translate(dx / 2, dy / 2)

        page = self.temp_doc.get_page(page_num)
        page.render_for_printing(cairo_ctx)

    def run(self):
        result = super().run(Gtk.PrintOperationAction.PRINT_DIALOG, self.app.window)
        if result == Gtk.PrintOperationResult.ERROR:
            dialog = Gtk.MessageDialog(
                self.app.window,
                0,
                Gtk.MessageType.ERROR,
                Gtk.ButtonsType.CLOSE,
                self.get_error(),
            )
            dialog.run()
            dialog.destroy()



# Copyright (C) 2020 pdfarranger contributors
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

__all__ = [
    "img2pdf_supported_img",
    "Page",
    "PageAdder",
    "PDFDoc",
    "PDFDocError",
    "PDFRenderer",
]

import sys
import os
import traceback
import mimetypes
import copy
import pathlib
import shutil
import tempfile
import threading
import time
import packaging.version as version
from typing import NamedTuple, Optional, Tuple, Union
import gettext
import gi
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk

gi.require_version("Poppler", "0.18")
from gi.repository import Poppler  # for the rendering of pdf pages
import cairo
from math import pi


try:
    import img2pdf

    img2pdf.Image.init()
    img2pdf_supported_img = [
        i for i in img2pdf.Image.MIME.values() if i.split("/")[0] == "image"
    ]
except ImportError:
    img2pdf_supported_img = []
    img2pdf = None


_ = gettext.gettext

Numeric = Union[float, int]


class Sides(NamedTuple):
    left: Numeric = 0
    right: Numeric = 0
    top: Numeric = 0
    bottom: Numeric = 0

    def __neg__(self) -> "Sides":
        """
        Pointwise unary minus

        Example:

        >>> -Sides(9, 3, 12, 6)
        Sides(left=-9, right=-3, top=-12, bottom=-6)
        """
        return Sides(*(-self[i] for i in range(4)))

    def __add__(self, other: Union["Sides", Numeric]) -> "Sides":
        """
        Pointwise addition

        Example:

        >>> Sides(9, 3, 12, 6) + Sides(1, 2, 3, 4)
        Sides(left=10, right=5, top=15, bottom=10)
        >>> Sides(9, 3, 12, 6) + 1
        Sides(left=10, right=4, top=13, bottom=7)
        """
        if isinstance(other, Sides):
            return Sides(*(self[i] + other[i] for i in range(4)))
        else:
            return Sides(*(self[i] + other for i in range(4)))

    def __sub__(self, other: Union["Sides", Numeric]) -> "Sides":
        """
        Pointwise subtraction

        Example:

        >>> Sides(9, 3, 12, 6) - Sides(1, 2, 3, 4)
        Sides(left=8, right=1, top=9, bottom=2)
        >>> Sides(9, 3, 12, 6) - 3
        Sides(left=6, right=0, top=9, bottom=3)
        """
        if isinstance(other, Sides):
            return Sides(*(self[i] - other[i] for i in range(4)))
        else:
            return Sides(*(self[i] -+ other for i in range(4)))

    def __mul__(self, other: Union["Sides", Numeric]) -> "Sides":
        """
        Pointwise multiplication

        Example:

        >>> Sides(9, 3, 12, 6) * Sides(1, 2, 3, 4)
        Sides(left=9, right=6, top=36, bottom=24)
        >>> Sides(9, 3, 12, 6) * 3
        Sides(left=27, right=9, top=36, bottom=18)
        """
        if isinstance(other, Sides):
            return Sides(*(self[i] * other[i] for i in range(4)))
        else:
            return Sides(*(self[i] * other for i in range(4)))

    def __truediv__(self, other: Union["Sides", Numeric]) -> "Sides":
        """
        Pointwise division

        Example:

        >>> Sides(9, 3, 12, 6) / Sides(1, 2, 3, 4)
        Sides(left=9.0, right=1.5, top=4.0, bottom=1.5)
        >>> Sides(9, 3, 12, 6) / 3
        Sides(left=3.0, right=1.0, top=4.0, bottom=2.0)
        """
        if isinstance(other, Sides):
            return Sides(*(self[i] / other[i] for i in range(4)))
        else:
            return Sides(*(self[i] / other for i in range(4)))

    def rotated(self, times: int) -> "Sides":
        """
        Rotate 90 degrees counter-clockwise 'times' times

        Examples:

        >>> Sides(9,3,12,6).rotated(1)
        Sides(left=12, right=6, top=3, bottom=9)
        >>> Sides(9,3,12,6).rotated(-3) == Sides(9,3,12,6).rotated(1)
        True
        """
        perm = (0, 2, 1, 3)
        return Sides(*(self[perm[(x + times) % 4]] for x in perm))


class Dims(NamedTuple):
    width: Numeric
    height: Numeric

    def __neg__(self) -> "Dims":
        """
        Pointwise unary minus

        Example:

        >>> -Dims(612, 792)
        Dims(width=-612, height=-792)
        """
        return Dims(*(-self[i] for i in range(2)))

    def __add__(self, other: Union["Dims", Numeric]) -> "Dims":
        """
        Pointwise addition

        Example:

        >>> Dims(612, 792) + Dims(612, 792)
        Dims(width=1224, height=1584)
        >>> Dims(612, 792) + 100
        Dims(width=712, height=892)
        """
        if isinstance(other, Dims):
            return Dims(*(self[i] + other[i] for i in range(2)))
        else:
            return Dims(*(self[i] + other for i in range(2)))

    def __sub__(self, other: Union["Dims", Numeric]) -> "Dims":
        """
        Pointwise subtraction

        Example:

        >>> Dims(612, 792) - Dims(306, 396)
        Dims(width=306, height=396)
        >>> Dims(612, 792) - 100
        Dims(width=512, height=692)
        """
        if isinstance(other, Dims):
            return Dims(*(self[i] - other[i] for i in range(2)))
        else:
            return Dims(*(self[i] -+ other for i in range(2)))

    def __mul__(self, other: Union["Dims", Numeric]) -> "Dims":
        """
        Pointwise multiplication

        Example:

        >>> Dims(612, 792) * Dims(0.5, 0.25)
        Dims(width=306.0, height=198.0)
        >>> Dims(612, 792) * 2
        Dims(width=1224, height=1584)
        """
        if isinstance(other, Dims):
            return Dims(*(self[i] * other[i] for i in range(2)))
        else:
            return Dims(*(self[i] * other for i in range(2)))

    def __truediv__(self, other: Union["Dims", Numeric]) -> "Dims":
        """
        Pointwise division

        Example:

        >>> Dims(612, 792) / Dims(2, 4)
        Dims(width=306.0, height=198.0)
        >>> Dims(612, 792) / 2
        Dims(width=306.0, height=396.0)
        """
        if isinstance(other, Dims):
            return Dims(*(self[i] / other[i] for i in range(2)))
        else:
            return Dims(*(self[i] / other for i in range(2)))

    def flipped(self) -> "Dims":
        """Swap height and width"""
        return Dims(self.height, self.width)

    def scaled(self, factor: float) -> "Dims":
        """Scale by factor"""
        return Dims(self.width * factor, self.height * factor)

    def int_scaled(self, factor: float) -> "Dims":
        """Scale by factor and round to nearest int"""
        return Dims(int(self.width * factor + 0.5), int(self.height * factor + 0.5))

    def cropped(self, crop: Sides) -> "Dims":
        """Crop using crop array"""
        return Dims(self.width * (1 - crop.left - crop.right), self.height * (1 - crop.top - crop.bottom))


class BasePage:
    """Common base class for Page and LayerPage"""

    def __init__(self, nfile, npage, copyname, angle, scale, crop: Sides, size_orig: Dims):
        self.nfile = nfile
        """The ID (from 1 to n) of the PDF file owning the page"""
        self.npage = npage
        """The ID (from 1 to n) of the page in its owner PDF document"""
        self.copyname = copyname
        """Filepath to the temporary stored file"""
        self.angle = angle
        self.scale = scale
        self.crop = crop
        """Left, right, top, bottom crop"""
        self.size_orig = size_orig
        """Width and height of the original page"""
        self.size = size_orig if angle in [0, 180] else size_orig.flipped()
        """Width and height"""

    def width_in_points(self) -> Numeric:
        """Return the page width in PDF points."""
        return self.size_in_points().width

    def height_in_points(self) -> Numeric:
        """Return the page height in PDF points."""
        return self.size_in_points().height

    def size_in_points(self) -> Dims:
        """Return the page size in PDF points."""
        return self.size.scaled(self.scale).cropped(self.crop)

    def size_in_mm(self) -> Dims:
        """Return the page size in mm."""
        return self.size_in_points() * 25.4 / 72

    def width_in_pixel(self):
        return self.size_in_pixel().width

    def height_in_pixel(self):
        return self.size_in_pixel().height

    def size_in_pixel(self):
        return self.size_in_points().int_scaled(self.zoom)

    @staticmethod
    def rotate_times(angle: int) -> int:
        """Convert an angle in degree to a number of 90° rotation (integer)."""
        return round((-angle  / 90) % 4)


class Page(BasePage):
    def __init__(self, nfile, npage, zoom, copyname, angle, scale, crop: Sides, hide: Sides, size_orig: Dims, description, layerpages):
        super().__init__(nfile, npage, copyname, angle, scale, Sides(*crop), size_orig)
        self.zoom = zoom
        self.hide = Sides(*hide)
        """Left, right, top, bottom hide"""
        self.thumbnail = None
        self.resample = -1
        self.preview = None
        """A low resolution thumbnail"""
        self.description = description
        """The text under the thumbnail"""
        self.layerpages = list(layerpages)

    def __repr__(self):
        return (f"Page({self.nfile}, {self.npage}, {self.zoom}, '{self.copyname}', "
                f"{self.angle}, {self.scale}, {self.crop}, {self.hide}, "
                f"{self.size_orig}, '{self.description}', {self.layerpages})")

    def rotate(self, angle: int):
        rt = self.rotate_times(angle)
        if rt == 0:
            return False
        self.crop = self.crop.rotated(rt)
        self.hide = self.hide.rotated(rt)
        self.angle = (self.angle + int(angle)) % 360
        self.size = self.size_orig if self.angle in [0, 180] else self.size_orig.flipped()
        for lp in self.layerpages:
            lp.rotate(rt)
        return True

    def unmodified(self):
        u = (self.angle == 0 and self.crop == Sides() and self.hide == Sides() and
             self.scale == 1 and len(self.layerpages) == 0)
        return u

    def serialize(self):
        """Convert to string for copy/past operations."""
        lpdata = [lp.serialize() for lp in self.layerpages]
        ts = [self.copyname, self.npage, self.description, self.angle, self.scale]
        ts += list(self.crop) + list(self.hide) + list(lpdata)
        return "///".join([str(v) for v in ts])

    def duplicate(self, incl_thumbnail=True):
        r = copy.copy(self)
        r.layerpages = [lp.duplicate() for lp in r.layerpages]
        if incl_thumbnail == False:
            del r.thumbnail  # to save ram
            r.thumbnail = None
            r.preview = None
        return r

    def split(self, vcrops, hcrops):
        """Split this page into a grid and return all but the top-left page."""
        newpages = []
        left, right, top, bottom = self.crop
        # If the page is cropped, adjust the new crop for the visible part of the page.
        hscale = 1 - (left + right)
        vscale = 1 - (top + bottom)
        vcrops = [(l * hscale, r * hscale) for (l, r) in vcrops]
        hcrops = [ (t * vscale, b * vscale) for (t, b) in hcrops]

        for (t, b) in reversed(hcrops):
            topcrop = top + t
            row_height = b - t
            bottomcrop = 1 - (topcrop + row_height)
            for (l, r) in reversed(vcrops):
                leftcrop = left + l
                col_width = r - l
                rightcrop = 1 - (leftcrop + col_width)
                crop = Sides(leftcrop, rightcrop, topcrop, bottomcrop)
                if l == 0.0 and t == 0.0:
                    # Update the original page
                    self.crop = crop
                else:
                    # Create a new cropped page
                    new = self.duplicate()
                    new.crop = crop
                    newpages.append(new)
        return newpages


class LayerPage(BasePage):
    """Page added as overlay or underlay on a Page."""

    def __init__(self, nfile, npage, copyname, angle, scale, crop, offset, laypos, size_orig: Dims):
        super().__init__(nfile, npage, copyname, angle, scale, Sides(*crop), size_orig)
        self.offset = Sides(*offset)
        """Left, right, top, bottom offset from dest page edges"""
        self.laypos = laypos
        """OVERLAY or UNDERLAY"""

    def __repr__(self):
        return (f"LayerPage({self.nfile}, {self.npage}, '{self.copyname}', {self.angle}, "
                f"{self.scale}, {self.crop}, {self.offset}, '{self.laypos}', {self.size_orig})")

    def rotate(self, times: int):
        if times != 0:
            self.crop = self.crop.rotated(times)
            self.offset = self.offset.rotated(times)
            self.angle = (self.angle - 90 * times) % 360
            self.size = self.size if times % 2 == 0 else self.size.flipped()

    def serialize(self):
        """Convert to string for copy/past operations."""
        ts = [self.copyname, self.npage, self.angle, self.scale, self.laypos]
        ts += list(self.crop) + list(self.offset)
        return "///".join([str(v) for v in ts])

    def duplicate(self):
        r = copy.copy(self)
        return r


class PDFDocError(Exception):
    def __init__(self, message):
        self.message = message


class _UnknownPasswordException(Exception):
    pass


class PasswordDialog(Gtk.Dialog):
    def __init__(self, parent, filename):
        super().__init__(
            title=_("Password required"),
            parent=parent,
            flags=Gtk.DialogFlags.MODAL,
            buttons=(
                "_Cancel",
                Gtk.ResponseType.CANCEL,
                "_OK",
                Gtk.ResponseType.OK,
            ),
        )
        self.set_default_response(Gtk.ResponseType.OK)
        bottommsg = _("The password will be remembered until you close PDF Arranger.")
        topmsg = _("The document “{}” is locked and requires a password before it can be opened.")
        label = Gtk.Label(label=topmsg.format(filename))
        label.set_max_width_chars(len(bottommsg)-6)
        label.set_line_wrap(True)
        label.set_size_request(0, -1)
        self.vbox.pack_start(label, False, False, 12)
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.entry = Gtk.Entry()
        self.entry.set_visibility(False)
        self.entry.set_activates_default(True)
        box.pack_start(Gtk.Label(label=_("Password")), False, False, 6)
        box.pack_start(self.entry, True, True, 6)
        self.vbox.pack_start(box, True, True, 0)
        self.vbox.pack_start(Gtk.Label(label=bottommsg), False, False, 12)
        self.set_resizable(False)

    def get_password(self):
        self.show_all()
        r = self.run()
        t = self.entry.props.text
        self.destroy()
        if r == Gtk.ResponseType.OK:
            return t
        else:
            raise _UnknownPasswordException()


def _img_to_pdf(images, tmp_dir, page_size=None):
    """Wrap img2pdf.convert to handle some corner cases"""
    if version.parse(img2pdf.__version__) < version.Version('0.4.2'):
        for num, image in enumerate(images):
            if isinstance(image, img2pdf.BytesIO):
                # Images from ImageExporter does not have transparency
                continue
            if isinstance(image, str):
                img = img2pdf.Image.open(image)
            else:
                img = img2pdf.Image.open(img2pdf.BytesIO(image))
            if (img.mode == "LA") or (img.mode != "RGBA" and "transparency" in img.info):
                # TODO: Find a way to keep image in P or L format and remove transparency.
                # This will work but converting from 1, L, P to RGB is not optimal.
                img = img.convert("RGBA")
            if img.mode == "RGBA":
                # Remove transparency as old img2pdf doesn't support it
                bg = img2pdf.Image.new("RGB", img.size, (255, 255, 255))
                bg.paste(img, mask=img.split()[-1])
                imgio = img2pdf.BytesIO()
                bg.save(imgio, "PNG")
                imgio.seek(0)
                images[num] = imgio
    try:
        # Try to handle invalid EXIF rotation
        rot = img2pdf.Rotation.ifvalid
    except AttributeError:
        # img2pdf is too old so we can't support invalid EXIF rotation
        rot = None
    kwargs = dict(rotation=rot)
    if page_size is not None:
        kwargs['layout_fun'] = img2pdf.get_layout_fun(page_size)
    try:
        pdf = img2pdf.convert(images, **kwargs)
    except ValueError as e:
        # Too small or large image
        raise PDFDocError(e)
    if tmp_dir is None:
        pdf_file = img2pdf.BytesIO()
        pdf_file.write(pdf)
    else:
        fd, pdf_file = tempfile.mkstemp(suffix=".pdf", dir=tmp_dir)
        os.close(fd)
        with open(pdf_file, "wb") as f:
            f.write(pdf)
    return pdf_file


class PDFDoc:
    """Class handling PDF documents."""

    def __from_file(self, parent, basename):
        uri = pathlib.Path(self.copyname).as_uri()
        askpass = False
        while True:
            try:
                if askpass:
                    self.password = PasswordDialog(parent, basename).get_password()
                self.document = Poppler.Document.new_from_file(uri, self.password)
                # When there is no encryption Poppler want None as password
                # while PikePDF want an empty string
                self.password = "" if self.password is None else self.password
                return
            except GLib.Error as e:
                askpass = e.message == "Document is encrypted"
                if not askpass:
                    raise e

    def __init__(self, filename, description, blank_size, stat, tmp_dir, parent):
        self.render_lock = threading.Lock()
        self.filename = os.path.abspath(filename)
        self.stat = stat
        if description is None:  # When importing files
            self.basename = os.path.basename(filename)
        else:  # When copy-pasting
            self.basename = description.split('\n')[0]
        self.blank_size = blank_size  # != None if page is blank
        self.password = ""
        filemime = mimetypes.guess_type(self.filename, strict=False)[0]
        if not filemime:
            raise PDFDocError(_("Unknown file format"))
        if filemime == "application/pdf":
            if self.filename.startswith(tmp_dir) and description is None:
                # In the "Insert Blank Page" we don't need to copy self.filename
                self.copyname = self.filename
                self.basename = ""
            else:
                fd, self.copyname = tempfile.mkstemp(suffix=".pdf", dir=tmp_dir)
                os.close(fd)
                shutil.copy(self.filename, self.copyname)
            try:
                self.__from_file(parent, self.basename)
            except GLib.Error as e:
                raise PDFDocError(e.message + ": " + filename)
        elif filemime.split("/")[0] == "image":
            if not img2pdf:
                raise PDFDocError(_("Image files are only supported with img2pdf"))
            if mimetypes.guess_type(filename, strict=False)[0] in img2pdf_supported_img:
                self.copyname = _img_to_pdf([filename], tmp_dir)
                uri = pathlib.Path(self.copyname).as_uri()
                self.document = Poppler.Document.new_from_file(uri, None)
            else:
                raise PDFDocError(_("Image format is not supported by img2pdf"))
            if filename.startswith(tmp_dir) and filename.endswith(".png"):
                os.remove(filename)
                self.basename = _("Clipboard image")
        else:
            raise PDFDocError(_("File is neither pdf nor image"))

        self.transparent_link_annots_removed = [False] * self.document.get_n_pages()

    def get_page(self, n_page):
        """Get a page where transparent link annotations are removed.

        By removing them memory usage will be lower.
        """
        page = self.document.get_page(n_page)
        if self.transparent_link_annots_removed[n_page]:
            return page
        annot_mapping_list = page.get_annot_mapping()
        for annot_mapping in annot_mapping_list:
            a = annot_mapping.annot
            if a.get_annot_type() == Poppler.AnnotType.LINK and a.get_color() is None:
                page.remove_annot(a)
        self.transparent_link_annots_removed[n_page] = True
        return page

class PageAdder:
    """Helper class to add pages to the current model."""

    def __init__(self, app):
        #: A PdfArranger instance
        self.app = app
        #: The pages which will be added by the commit method
        self.pages = []
        #: Where to insert pages relatively to treerowref
        self.before = False
        #: Where to insert pages. If None pages are inserted at the end
        self.treerowref = None
        self.stat_cache = {}
        self.content = []
        self.pdfqueue_used = True

    def move(self, treerowref, before):
        """Insert pages at the given location."""
        self.before = before
        self.treerowref = treerowref

    def get_pdfdoc(self, filename: str, description: Optional[str] = None, blank_size=None) -> Optional[Tuple[PDFDoc, int, bool]]:
        """Get the pdfdoc object for the filename.

        pdfqueue is searched for the filename. If it is not found a pdfdoc is created
        and added to pdfqueue.
        Returns: pdfdoc object, it's file number, if a new pdfdoc was created.
        """
        for i, it_pdfdoc in enumerate(self.app.pdfqueue):
            if filename == it_pdfdoc.copyname:
                # File of copy-pasted page was found in pdfqueue.
                # Files in tmp_dir are never modified by the app and are not expected
                # to be modified by the user either -> files are equal if names match.
                return it_pdfdoc, i + 1, False

        if not filename in self.stat_cache:
            try:
                s = os.stat(filename)
                self.stat_cache[filename] = s.st_dev, s.st_ino, s.st_mtime
            except OSError as e:
                print(traceback.format_exc())
                self.app.error_message_dialog(e)
                return None
        for i, it_pdfdoc in enumerate(self.app.pdfqueue):
            if self.stat_cache[filename] == it_pdfdoc.stat:
                # Imported file was found in pdfqueue
                return it_pdfdoc, i + 1, False

        try:
            pdfdoc = PDFDoc(filename, description, blank_size, self.stat_cache[filename],
                            self.app.tmp_dir, self.app.window)
        except _UnknownPasswordException:
            return None
        except PDFDocError as e:
            print(e.message, file=sys.stderr)
            self.app.error_message_dialog(e.message)
            return None

        self.app.pdfqueue.append(pdfdoc)
        return pdfdoc, len(self.app.pdfqueue), True

    def get_layerpages(self, layerdata):
        """Create LayerPage objects from layerdata."""
        layerpages = []
        if layerdata is None:
            return layerpages
        for filename, npage, angle, scale, laypos, crop, offset in layerdata:
            doc_data = self.get_pdfdoc(filename)
            if doc_data is None:
                return None
            pdfdoc, nfile, _ = doc_data
            copyname = pdfdoc.copyname
            size = Dims(*pdfdoc.get_page(npage - 1).get_size())
            ld = nfile, npage, copyname, angle, scale, crop, offset, laypos, size
            layerpages.append(LayerPage(*ld))
        return layerpages

    def addpages(self, filename, page=-1, description=None, angle=0, scale=1.0, crop=Sides(0, 0, 0, 0), hide=Sides(0, 0, 0, 0), layerdata=None):
        c = 'pdf' if page == -1 and os.path.splitext(filename)[1].lower() == '.pdf' else 'other'
        self.content.append(c)
        self.pdfqueue_used = len(self.app.pdfqueue) > 0

        doc_data = self.get_pdfdoc(filename, description)
        if doc_data is None:
            return
        pdfdoc, nfile, doc_added = doc_data

        if (doc_added and pdfdoc.copyname != pdfdoc.filename and description is None and not
                (filename.startswith(self.app.tmp_dir) and filename.endswith(".png"))):
            self.app.import_directory = os.path.split(filename)[0]
            self.app.export_directory = self.app.import_directory

        n_end = pdfdoc.document.get_n_pages()
        n_start = min(n_end, max(1, page))
        if page != -1:
            n_end = max(n_start, min(n_end, page))

        layerpages = self.get_layerpages(layerdata)
        if layerpages is None:
            return

        for npage in range(n_start, n_end + 1):
            page = pdfdoc.document.get_page(npage - 1)
            if description is None:
                shortname = os.path.splitext(pdfdoc.basename)[0]
                desc = "".join([shortname, "\n", _("page"), " ", str(npage)])
            else:
                desc = description
            self.pages.append(
                Page(
                    nfile,
                    npage,
                    self.app.zoom_scale,
                    pdfdoc.copyname,
                    angle,
                    scale,
                    crop,
                    hide,
                    Dims(*page.get_size()),
                    desc,
                    layerpages,
                )
            )

    def commit(self, select_added, add_to_undomanager):
        if len(self.pages) == 0:
            return False
        if add_to_undomanager:
            self.app.undomanager.commit("Add")
            if self.pdfqueue_used or len(self.content) > 1 or self.content[0] != 'pdf':
                self.app.set_unsaved(True)
            self.content = []
        if not self.before and self.treerowref:
            self.pages.reverse()
        with self.app.render_lock():
            for p in self.pages:
                m = [p, p.description]
                if self.treerowref:
                    iter_to = self.app.model.get_iter(self.treerowref.get_path())
                    if self.before:
                        it = self.app.model.insert_before(iter_to, m)
                    else:
                        it = self.app.model.insert_after(iter_to, m)
                else:
                    it = self.app.model.append(m)
                if select_added:
                    path = self.app.model.get_path(it)
                    self.app.iconview.select_path(path)
        if add_to_undomanager:
            self.app.update_iconview_geometry()
            GObject.idle_add(self.app.retitle)
            self.app.update_max_zoom_level()
            self.app.silent_render()
            self.app.update_statusbar()
            self.scroll()
        self.pages = []
        return True

    def scroll(self):
        """Scroll to first added page."""
        if len(self.app.model) - len(self.pages) == 0:
            return
        if self.treerowref:
            iref = self.treerowref.get_path().get_indices()[0]
        else:
            iref = len(self.app.model) - 1 - len(self.pages)
            self.before = False
        if self.before:
            scroll_path = Gtk.TreePath.new_from_indices([max(iref - len(self.pages), 0)])
        else:
            scroll_path = Gtk.TreePath.new_from_indices([max(iref + 1, 0)])
        self.app.iconview.scroll_to_path(scroll_path, False, 0, 0)


class PDFRenderer(threading.Thread, GObject.GObject):
    def __init__(self, model, pdfqueue, visible_range, columns_nr, max_nqueue=-1):
        threading.Thread.__init__(self)
        GObject.GObject.__init__(self)
        self.model = model
        self.pdfqueue = pdfqueue
        self.visible_start = visible_range[0]
        self.visible_end = visible_range[1]
        self.columns_nr = columns_nr
        self.max_nqueue = max_nqueue
        self.nqueue = 0
        self.mem_usage = 0
        self.model_lock = threading.Lock()
        self.quit = False

    def run(self):
        """Render thumbnails and less memory consuming previews.

        Thumbnails are rendered for the visible range and its near area. Memory usage is estimated
        and if it goes too high distant thumbnails are replaced with previews. Previews will be
        rendered for all pages.
        """
        for num in range(self.visible_start, self.visible_end + 1):
            if self.quit:
                return
            with self.model_lock:
                if not 0 <= num < len(self.model):
                    break
                path = Gtk.TreePath.new_from_indices([num])
                ref = Gtk.TreeRowReference.new(self.model, path)
                p = self.model[path][0].duplicate()
            if p.resample != 1 / p.zoom:
                self.update(p, ref, p.zoom, False)
        mem_limit = False
        for off in range(1, len(self.model)):
            for num in self.visible_end + off, self.visible_start - off:
                if self.quit:
                    return
                with self.model_lock:
                    if not 0 <= num < len(self.model):
                        continue
                    path = Gtk.TreePath.new_from_indices([num])
                    ref = Gtk.TreeRowReference.new(self.model, path)
                    p = self.model[path][0].duplicate()
                if off <= self.columns_nr * 5:
                    # Thumbnail
                    zoom = p.zoom
                    is_preview = False
                elif mem_limit or p.resample < 0:
                    # Preview. Always render to about 4000 pixels = about 16kb
                    zoom = (1 / p.scale) * (4000 / (p.size[0] * p.size[1])) ** .5
                    is_preview = True
                else:
                    # Thumbnail is distant and total mem usage is small
                    # -> don't update thumbnail, just take memory usage into account
                    zoom = 1 / p.resample
                if p.resample != 1 / zoom:
                    size = self.update(p, ref, zoom, is_preview)
                else:
                    size = p.thumbnail.get_width(), p.thumbnail.get_height()
                mem_limit = self.mem_at_limit(size)
        self.finish()

    def mem_at_limit(self, size):
        """Estimate memory usage of rendered thumbnails. Return True when mem_usage > mem_limit."""
        mem_limit = 300  # Mb (About. Size will depend on thumbnail content.)
        if self.mem_usage > mem_limit:
            return True
        self.mem_usage += size[0] * size[1] * 4 / (1024 * 1024)  # 4 byte/pixel
        return False

    def render(self, cr, p):
        if self.quit:
            return
        pdfdoc = self.pdfqueue[p.nfile - 1]
        page = pdfdoc.get_page(p.npage - 1)
        with pdfdoc.render_lock:
            page.render(cr)

    def update(self, p: Page, ref, zoom, is_preview):
        """Render and emit updated thumbnails."""
        if (is_preview and p.preview) and (p.resample != -1):
            # Reuse the preview if it exist, unless it is marked for re-render
            thumbnail = p.preview
        else:
            wpoi = p.size.width * (1 - p.crop.left - p.crop.right)
            hpoi = p.size.height * (1 - p.crop.top - p.crop.bottom)
            wpix = max(1, int(0.5 + wpoi * p.scale * zoom))
            hpix = max(1, int(0.5 + hpoi * p.scale * zoom))
            wpix0, hpix0 = (wpix, hpix) if p.angle in [0, 180] else (hpix, wpix)
            rotation = round((int(p.angle) % 360) / 90) * 90

            thumbnail = cairo.ImageSurface(cairo.FORMAT_ARGB32, wpix0, hpix0)
            cr = cairo.Context(thumbnail)
            if rotation > 0:
                cr.translate(wpix0 / 2, hpix0 / 2)
                cr.rotate(-rotation * pi / 180)
                cr.translate(-wpix / 2, -hpix / 2)
            cr.scale(wpix / wpoi, hpix / hpoi)
            cr.translate(-p.crop.left * p.size.width, -p.crop.top * p.size.height)
            self.add_layers(cr, p, layer='UNDERLAY')
            cr.save()
            if rotation > 0:
                cr.translate(*p.size.scaled(0.5))
                cr.rotate(rotation * pi / 180)
                cr.translate(*p.size_orig.scaled(-0.5))
            self.render(cr, p)
            cr.restore()
            self.add_layers(cr, p, layer='OVERLAY')

            if p.hide != Sides():
                cr.set_source_rgb(1, 1, 1)
                cr.rectangle(0, 0, p.size.width, p.size.height)
                x = p.size.width * p.hide.left
                y = p.size.height * p.hide.top
                w = p.size.width * (1 - p.hide.left - p.hide.right)
                h = p.size.height * (1 - p.hide.top - p.hide.bottom)
                cr.rectangle(x, y, w, h)
                cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
                cr.fill()

        if self.quit:
            return 0, 0
        if self.max_nqueue > 0:
            # Limit queue length for lower memory usage
            self.nqueue += 1
            while self.nqueue > self.max_nqueue:
                time.sleep(0.1)

        GObject.idle_add(
            self.emit,
            "update_thumbnail",
            ref,
            thumbnail,
            zoom,
            p.scale,
            is_preview,
            priority=GObject.PRIORITY_LOW,
        )
        return thumbnail.get_width(), thumbnail.get_height()

    def add_layers(self, cr, p: Page, layer):
        layerpages = p.layerpages if layer == 'OVERLAY' else reversed(p.layerpages)
        for lp in layerpages:
            if self.quit:
                return
            if layer != lp.laypos:
                continue
            cr.save()
            cr.translate(p.size.width * lp.offset.left, p.size.height * lp.offset.top)
            cr.scale(lp.scale / p.scale, lp.scale / p.scale)
            x = lp.size.width * lp.crop.left
            y = lp.size.height * lp.crop.top
            w = lp.size.width * (1 - lp.crop.left - lp.crop.right)
            h = lp.size.height * (1 - lp.crop.top - lp.crop.bottom)
            cr.translate(-x, -y)
            cr.rectangle(x, y, w, h)
            cr.clip()
            rotation = round((int(lp.angle) % 360) / 90) * 90
            if rotation > 0:
                cr.translate(*lp.size.scaled(0.5))
                cr.rotate(rotation * pi / 180)
                cr.translate(*lp.size_orig.scaled(-0.5))
            self.render(cr, lp)
            cr.restore()

    def finish(self):
        """Signal rendering ended (for statusbar and malloc_trim)."""
        GObject.idle_add(
            self.emit,
            "update_thumbnail",
            None,
            None,
            0,
            0,
            False,
            priority=GObject.PRIORITY_LOW,
        )



# Copyright (C) 2020 pdfarranger contributors
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gi.repository import Gtk
import gettext

_ = gettext.gettext

class Dialog(Gtk.Dialog):
    """ A dialog box to split pages into a grid of pages"""
    def __init__(self, window):
        super().__init__(
            title=_("Split Pages"),
            parent=window,
            flags=Gtk.DialogFlags.MODAL,
            buttons=(
                "_Cancel",
                Gtk.ResponseType.CANCEL,
                "_OK",
                Gtk.ResponseType.OK,
            ),
        )
        buttonbox = self.get_action_area()
        buttons = buttonbox.get_children()
        self.set_focus(buttons[1])
        self.set_resizable(False)
        self.split_count = {'vertical' : 2, 'horizontal' : 1}
        self.even_splits = {'vertical' : True, 'horizontal' : True}
        self.vmodel = Gtk.ListStore(int, int)
        self.hmodel = Gtk.ListStore(int, int)
        self.model = {'vertical' : self.vmodel, 'horizontal' : self.hmodel}
        self.vspin = Gtk.SpinButton()
        self.hspin = Gtk.SpinButton()
        self.spinbuttons = {'vertical' : self.vspin, 'horizontal' : self.hspin}
        self.vcheck = Gtk.CheckButton()
        self.hcheck = Gtk.CheckButton()
        self.checkbuttons = {'vertical' : self.vcheck, 'horizontal' : self.hcheck}

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.vbox.pack_start(hbox, True, True, 0)
        for direction in ['vertical', 'horizontal']:
            frame = self._build_frame(direction)
            hbox.pack_start(frame, True, True, 0)
        self.show_all()

    def _build_frame(self, direction):
        frame_txt = {'vertical' : _("Columns"), 'horizontal' : _("Rows")}
        label_txt = {'vertical' : _("Vertical Splits"), 'horizontal' : _("Horizontal Splits")}
        checkbutton_txt = {'vertical' : _("Equal column width"), 'horizontal' : _("Equal row height")}

        frame = Gtk.Frame(label=frame_txt[direction])
        frame.props.margin = 8
        frame.props.margin_bottom = 0
        grid = Gtk.Grid()
        frame.add(grid)
        label = Gtk.Label(label_txt[direction])
        label.set_alignment(0.0, 0.5)
        label.props.margin = 8
        label.props.margin_bottom = 6
        grid.attach(label, 0, 0, width=1, height=1)
        adjustment = Gtk.Adjustment(value=self.split_count[direction], lower=1, upper=10, step_incr=1)
        self.spinbuttons[direction].set_adjustment(adjustment)
        self.spinbuttons[direction].connect("value-changed", self._update_split, direction)
        grid.attach(self.spinbuttons[direction], 1, 0, width=1, height=1)
        self.checkbuttons[direction].set_label(checkbutton_txt[direction])
        self.checkbuttons[direction].set_active(True)
        self.checkbuttons[direction].connect("toggled", self._even_split_toggled, direction)
        grid.attach(self.checkbuttons[direction], 0, 1, width=2, height=1)
        treeview = self._build_model(direction)
        grid.attach(treeview, 0, 2, width=2, height=1)
        return frame

    def _build_model(self, direction):
        label1 = {'vertical' : _("#Col"), 'horizontal' : _("#Row")}
        label2 = {'vertical' : _("Width in %"), 'horizontal' : _("Height in %")}

        split_count = self.split_count[direction]
        for s in range(1, split_count + 1):
            self.model[direction].append([s, 100 // split_count])
        treeview = Gtk.TreeView(model=self.model[direction])
        cr = Gtk.CellRendererText()
        heading = Gtk.TreeViewColumn(label1[direction], cr, text=0)
        treeview.append_column(heading)
        cr = Gtk.CellRendererSpin()
        cr.connect("edited", self._edited, direction)
        cr.set_property("editable", True)
        adjustment = Gtk.Adjustment(value=100, lower=0, upper=100, step_increment=1)
        cr.set_property("adjustment", adjustment)
        heading = Gtk.TreeViewColumn(label2[direction], cr, text=1)
        treeview.append_column(heading)
        return treeview

    def _edited(self, widget, path, value, direction):
        # Avoid that a user bypasses the limits of the spin buttons
        # via keyboard input and pressing enter.
        value = min(max(0, int(value)), 100)
        if self.model[direction][path][1] == int(value):
            return
        if self.checkbuttons[direction].get_active():
            # Populate all entries with the new value
            # The sum may exceed 100, allowing for overlap.
            crop_sum = len(self.model[direction]) * value
            if crop_sum >= 100:
                for i in range(len(self.model[direction])):
                    self.model[direction][i][1] = value
        else:
            # Unevenly sized tiles. Fix the sum to match 100,
            # adjusting entries from bottom to top.
            self.model[direction][path][1] = value
            crop_sum = 0
            for i in range(len(self.model[direction])):
                crop_sum += self.model[direction][i][1]
            delta = 100 - crop_sum
            cur_idx = int(path)
            for i in reversed(range(len(self.model[direction]))):
                if i != cur_idx:
                    sign = delta/abs(delta)
                    s = sign * min(abs(delta), self.model[direction][i][1])
                    self.model[direction][i][1] += s
                    delta -= s
                    if delta == 0:
                        break

    def _update_split(self, _event, direction):
        self.split_count[direction] = self.spinbuttons[direction].get_value_as_int()
        if self.even_splits[direction]:
            self.model[direction].clear()
            # Partition evenly
            count = self.split_count[direction]
            frac = 100 // count
            partition = [frac] * (count - 1)
            partition.append(100 - (count - 1) * frac)
            for i, frac in enumerate(partition, start = 1):
                self.model[direction].append([i, frac])
        else:
            delta = self.split_count[direction] - len(self.model[direction])
            if delta > 0:
                # Add delta zero rows
                idx = len(self.model[direction]) + 1
                for i in range(delta):
                    self.model[direction].append([idx + i, 0])
            if delta < 0:
                # Delete the last delta entries and ensure that the sum is 100
                s = 0
                for i in range(abs(delta)):
                    s += self.model[direction][-1][1]
                    del self.model[direction][-1]
                self.model[direction][-1][1] += s

    def _even_split_toggled(self, button, direction):
        self.even_splits[direction] = button.get_active()
        self._update_split(None, direction)

    def run_get(self):
        result = self.run()
        vcrops = None
        hcrops = None
        if result == Gtk.ResponseType.OK:
            vcrops = _crops(self.model['vertical'])
            hcrops = _crops(self.model['horizontal'])
        self.destroy()
        return vcrops, hcrops

# Operates over a 1D-array of [PAGE_N, FRAC] arrays
def _crops(linear_tiles):
    # Convert the tile sizes into a list of tuples (start, end) such that
    # tile size = end - start.
    num_splits = len(linear_tiles)
    crops = [(0, 0.01 * linear_tiles[0][1])] * num_splits
    crop_sum = 0.01 * linear_tiles[0][1]
    for i in range(1, num_splits):
        size = 0.01 * linear_tiles[i][1]
        crop_sum += size
        crops[i] = (crops[i-1][1], crops[i-1][1] + size)

    # Remove zero-sized tiles
    crops = [t for t in crops if t[0] < t[1]]

    overlap = crop_sum - 1.0 # nonnegative
    if overlap == 0.0:
        return crops

    # We have multiple splits and crop_sum > 1.0
    # 35,35,35 => [0,35],[32.5,67.5],[65,100]; overlap = 5
    # 60,60 => [0,60],[40,100]; overlap = 20
    # In general:
    #   [start=last-overlap/(num_splits-1),end=start+size]
    overlap_per_tile = overlap / (num_splits - 1)

    # Overlap is only defined when all tiles have the same size.
    size = 0.01 * linear_tiles[0][1]
    crops[0] = (0, size)
    for i in range(1, num_splits - 1):
        start = crops[i-1][1] - overlap_per_tile
        end = start + size
        crops[i] = (start, end)
    crops[-1] = (1.0 - size, 1.0)
    return crops



# Copyright (C) 2020 pdfarranger contributors
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gi.repository import Gtk, Gdk, GObject
import gettext
import cairo
import locale

from math import pi

from .core import Sides, Dims, PDFRenderer

_ = gettext.gettext


def scale(model, selection, factor):
    """Set the scale factor of a selection of pages."""
    changed = False
    try:
        width, height = factor
    except TypeError:
        width, height = None, None
    for path in selection:
        it = model.get_iter(path)
        page = model.get_value(it, 0)
        page_size = page.size.cropped(page.crop)
        if width is None:
            f = factor
        else:
            # TODO: allow to change aspect ratio
            f = min(*Dims(width, height) / page_size)
        # Page size must be in [72, 14400] (PDF standard requirement)
        f = max(f, *(Dims(72, 72) / page_size))
        f = min(f, *(Dims(14400, 14400) / page_size))
        if page.scale != f:
            changed = True
        page.resample = page.resample * f / page.scale
        for lp in page.layerpages:
            lp.scale = lp.scale * f / page.scale
        page.scale = f
        model.set_value(it, 0, page)
    return changed


class _LinkedSpinButton(Gtk.SpinButton):
    """ A spin button which can be bound to an other button """

    def __init__(self, minval, maxval, step, page=None):
        Gtk.SpinButton.__init__(self)
        self.set_digits(20)
        self.set_width_chars(9)
        self.connect("output", self.__output)
        self.set_range(minval, maxval)
        self.set_increments(step, step if page is None else page)
        self.changing_from_brother = False

    def __output(self, _user_data):
        """ output signal handler to remove unneeded 0 """
        s = locale.format_string("%.8g", self.get_adjustment().get_value())
        self.get_buffer().set_text(s, len(s))
        return True


class _RadioStackSwitcher(Gtk.Box):
    """ Same as GtkStackSwitcher but with radio button (i.e different semantic) """

    def __init__(self, margin=10):
        super().__init__()
        self.props.orientation = Gtk.Orientation.VERTICAL
        self.set_spacing(margin)
        self.props.margin = margin
        self.radiogroup = []
        self.stack = Gtk.Stack()
        self.button_box = Gtk.Box()
        self.button_box.set_spacing(margin)
        self.add(self.button_box)
        self.add(self.stack)
        self.selected_child = None
        self.selected_name = None

    def add_named(self, child, name, title):
        radio = Gtk.RadioButton.new_with_label(None, title)
        if len(self.radiogroup) > 0:
            radio.join_group(self.radiogroup[-1])
        self.radiogroup.append(radio)
        radio.set_hexpand(True)
        radio.set_halign(Gtk.Align.CENTER)
        radio.connect("toggled", self.__radio_handler, name)
        self.button_box.add(radio)
        self.stack.add_named(child, name)
        if self.selected_child is None:
            self.selected_child = child
            self.selected_name = name

    def __radio_handler(self, button, name):
        if button.props.active:
            self.stack.set_visible_child_name(name)
            self.selected_name = name
            self.selected_child = self.stack.get_child_by_name(name)


class _RelativeScalingWidget(Gtk.Box):
    """ A form to specify the relative scaling factor """

    def __init__(self, current_scale, margin=10):
        super().__init__(valign=Gtk.Align.CENTER, halign=Gtk.Align.CENTER)
        self.props.spacing = margin
        self.add(Gtk.Label(label=_("Scale factor")))
        # Largest page size is 200 inch and smallest is 1 inch
        # so we can set a min and max
        self.entry = _LinkedSpinButton(100 / 200.0, 100 * 200.0, 1, 10)
        self.entry.set_activates_default(True)
        self.add(self.entry)
        self.add(Gtk.Label(label=_("%")))
        self.entry.set_value(current_scale * 100)

    def get_value(self):
        return self.entry.get_value() / 100


class PaperSizeWidget(Gtk.Grid):
    def __init__(self, size, margin=16):
        super().__init__(margin=margin, row_spacing=8, column_spacing=8, row_homogeneous=True)

        self.attach(Gtk.Label(label=_("Width"), halign=Gtk.Align.START), 1, 1, 1, 1)
        self.width_entry = _LinkedSpinButton(25.4, 5080, 1, 0)
        self.w_entry_id = self.width_entry.connect('value-changed', self.width_changed)
        self.attach(self.width_entry, 2, 1, 1, 1)
        self.attach(Gtk.Label(label=_("mm"), halign=Gtk.Align.START), 4, 1, 1, 1)

        self.attach(Gtk.Label(label=_("Height"), halign=Gtk.Align.START), 1, 2, 1, 1)
        self.height_entry = _LinkedSpinButton(25.4, 5080, 1, 10)
        self.h_entry_id = self.height_entry.connect('value-changed', self.height_changed)
        self.attach(self.height_entry, 2, 2, 1, 1)
        self.attach(Gtk.Label(label=_("mm"), halign=Gtk.Align.START), 4, 2, 1, 1)

        self.attach(Gtk.Label(_("Paper size"), halign=Gtk.Align.START), 1, 3, 1, 1)
        self.combo = Gtk.ComboBoxText(margin=0)
        self.combo_changed_id = self.combo.connect('changed', self.paper_size_changed)
        self.papers = [Gtk.PaperSize.new_custom('Custom', _("Custom"), 0, 0, Gtk.Unit.MM)]
        paper_list = ['iso_a3', 'iso_a4', 'iso_a5', 'na_letter', 'na_legal', 'na_ledger']
        self.papers += [Gtk.PaperSize.new(p) for p in paper_list]
        for p in self.papers:
            p.size = [round(p.get_width(Gtk.Unit.MM), 5), round(p.get_height(Gtk.Unit.MM), 5)]
            self.combo.append(None, p.get_display_name())
        self.attach(self.combo, 2, 3, 1, 1)

        self.attach(Gtk.Label(_("Orientation"), halign=Gtk.Align.START), 1, 4, 1, 1)
        self.port = Gtk.RadioButton(label=_("Portrait"), group=None)
        self.land = Gtk.RadioButton(label=_("Landscape"), group=self.port)
        self.port.connect('clicked', self.orientation_clicked)
        box1 = Gtk.Box()
        box1.pack_start(self.port, True, True, 0)
        box1.pack_start(self.land, True, True, 0)
        self.attach(box1, 2, 4, 1, 1)

        box2 = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box2.pack_start(Gtk.Label("┐"), True, True, 0)
        self.ratio_cb = Gtk.CheckButton(margin_top=1)
        box2.pack_start(self.ratio_cb, True, True, 0)
        box2.pack_start(Gtk.Label("┘"), True, True, 0)
        self.attach(box2, 3, 1, 1, 2)

        if size is None:
            size = [210, 297]  # A4 by default
            self.ratio_cb.set_sensitive(False)
        else:
            self.ratio_cb.set_active(True)
            self.ratio_cb.connect('clicked', self.width_changed)

        self.ratios = [size[0] / size[1], size[1] / size[0]]
        self.set_entry_size(size)
        self.select_paper_and_orientation()
        self.update_entry_limits()
        self.ratio_cb.connect('clicked', self.update_entry_limits)

    def update_entry_limits(self, _widget=None):
        r = self.ratios if self.ratio_cb.get_active() else [1, 1]
        wrange = max(25.4, 25.4 * r[0]), min(5080, 5080 * r[0])
        hrange = max(25.4, 25.4 * r[1]), min(5080, 5080 * r[1])
        self.width_entry.set_range(*wrange)
        self.height_entry.set_range(*hrange)

    def width_changed(self, _widget):
        if self.ratio_cb.get_active():
            width = self.width_entry.get_value()
            self.set_entry_size((width, width / self.ratios[0]))
        self.select_paper_and_orientation()

    def height_changed(self, _widget):
        if self.ratio_cb.get_active():
            height = self.height_entry.get_value()
            self.set_entry_size((height / self.ratios[1], height))
        self.select_paper_and_orientation()

    def orientation_clicked(self, _widget):
        size = self.get_value(Gtk.Unit.MM)
        self.width_entry.set_range(25.4, 5080)
        self.height_entry.set_range(25.4, 5080)
        self.set_entry_size(sorted(size, reverse=self.land.get_active()))
        self.ratios.sort(reverse=self.land.get_active())
        self.update_entry_limits()

    def paper_size_changed(self, combo):
        paper = self.papers[combo.get_active()]
        size = paper.get_width(Gtk.Unit.MM), paper.get_height(Gtk.Unit.MM)
        self.set_entry_size(sorted(size, reverse=self.land.get_active()))
        if round(size[0] / size[1], 5) not in [round(r, 5) for r in self.ratios]:
            self.ratio_cb.set_active(False)
            self.update_entry_limits()

    def select_paper_and_orientation(self):
        size = self.get_value(Gtk.Unit.MM)
        self.papers[0].set_size(*size, Gtk.Unit.MM)
        size = [round(s, 5) for s in size]
        for num, paper in reversed(list(enumerate(self.papers))):
            if paper.size in [size, size[::-1]]:
                break
        self.combo.set_active(num)
        self.port.set_active(size[0] < size[1])
        self.land.set_active(size[0] > size[1])

    def set_entry_size(self, size):
        """Set entry size in mm"""
        with GObject.signal_handler_block(self.width_entry, self.w_entry_id):
            self.width_entry.set_value(size[0])
        with GObject.signal_handler_block(self.height_entry, self.h_entry_id):
            self.height_entry.set_value(size[1])

    def get_value(self, unit=Gtk.Unit.POINTS):
        """Get entry size in points or mm"""
        size = [self.width_entry.get_value(), self.height_entry.get_value()]
        if unit == Gtk.Unit.POINTS:
            size = [s * 72 / 25.4 for s in size]
        return size


class _CropHideWidget(Gtk.Frame):
    sides = ('L', 'R', 'T', 'B')
    side_names = {'L': _('Left'), 'R': _('Right'), 'T': _('Top'), 'B': _('Bottom')}
    opposite_sides = {'L': 'R', 'R': 'L', 'T': 'B', 'B': 'T'}

    def __init__(self, val, margin=12):
        super().__init__(shadow_type=Gtk.ShadowType.NONE)
        grid = Gtk.Grid(halign=Gtk.Align.CENTER)
        grid.set_column_spacing(margin)
        grid.set_row_spacing(margin)
        grid.props.margin = margin
        self.add(grid)
        label = Gtk.Label(
            label=_(
                'Cropping/hiding does not remove any content '
                'from the PDF file, it only hides it.'
            )
        )
        label.props.margin = margin
        label.set_line_wrap(True)
        label.set_max_width_chars(38)
        grid.attach(label, 0, 0, 3, 1)
        self.spin_changed_callback = None
        self.spin_list = []
        units = 2 * [_('% of width')] + 2 * [_('% of height')]

        for row, side in enumerate(_CropHideWidget.sides):
            label = Gtk.Label(label=_CropHideWidget.side_names[side])
            label.set_alignment(0.0, 0.5)
            grid.attach(label, 0, row + 1, 1, 1)

            adj = Gtk.Adjustment(
                value=100.0 * val.pop(0),
                lower=0.0,
                upper=90.0,
                step_increment=1.0,
                page_increment=5.0,
                page_size=0.0,
            )
            spin = Gtk.SpinButton(adjustment=adj, climb_rate=0, digits=1)
            spin.set_activates_default(True)
            spin.connect('value-changed', self.__set_value, self, side)
            self.spin_list.append(spin)
            grid.attach(spin, 1, row + 1, 1, 1)

            label = Gtk.Label(label=units.pop(0))
            label.set_alignment(0.0, 0.5)
            grid.attach(label, 2, row + 1, 1, 1)

    @staticmethod
    def __set_value(spinbutton, self, side):
        opp_side = self.opposite_sides[side]
        adj = self.spin_list[self.sides.index(opp_side)].get_adjustment()
        limit = 90.0 - spinbutton.get_value()
        adj.set_upper(limit)
        opp_spinner = self.spin_list[self.sides.index(opp_side)]
        opp_spinner.set_value(min(opp_spinner.get_value(), limit))
        if callable(self.spin_changed_callback):
            self.spin_changed_callback()

    def set_spinb_changed_callback(self, callback):
        self.spin_changed_callback = callback

    def set_val(self, val):
        for i, spin in enumerate(self.spin_list):
            spin.set_value(val[i] * 100)

    def get_val(self):
        return Sides(*(spin.get_value() / 100.0 for spin in self.spin_list))


class BaseDialog(Gtk.Dialog):
    def __init__(self, title, parent, prepend_buttons=None):
        buttons = () if prepend_buttons is None else prepend_buttons
        buttons += (
            _("_Cancel"), Gtk.ResponseType.CANCEL,
            _("_OK"), Gtk.ResponseType.OK,
        )
        super().__init__(
            title=title,
            parent=parent,
            flags=Gtk.DialogFlags.MODAL,
            buttons=buttons,
        )
        self.set_default_response(Gtk.ResponseType.OK)


class ScaleDialog(BaseDialog):
    """ A dialog box to define page size or scale factor """

    def __init__(self, model, selection, window):
        super().__init__(title=_("Page size"), parent=window)
        self.set_resizable(False)
        page = model.get_value(model.get_iter(selection[-1]), 0)
        paper_widget = PaperSizeWidget(page.size_in_mm(), margin=1)
        paper_widget.attach(Gtk.Label(_("Fit mode"), halign=Gtk.Align.START), 1, 5, 1, 1)
        self.combo = Gtk.ComboBoxText()
        self.combo.append('SCALE', _("Scale"))
        self.combo.append('SCALE-ADD-MARG', _("Scale & Add margins"))
        self.combo.append('CROP-ADD-MARG', _("Crop & Add margins"))
        self.combo.set_active(0)
        paper_widget.attach(self.combo, 2, 5, 1, 1)
        rel_widget = _RelativeScalingWidget(page.scale)
        self.scale_stack = _RadioStackSwitcher(margin=15)
        self.scale_stack.add_named(paper_widget, "Fit", _("Fit to paper"))
        self.scale_stack.add_named(rel_widget, "Relative", _("Relative"))
        pagesizeframe = Gtk.Frame(shadow_type=Gtk.ShadowType.NONE)
        pagesizeframe.add(self.scale_stack)
        self.vbox.pack_start(pagesizeframe, True, True, 0)
        self.show_all()
        self.selection = selection

    def run_get(self):
        """ Open the dialog and return the size value """
        result = self.run()
        val = None
        if result == Gtk.ResponseType.OK:
            s = self.scale_stack
            mode = 'SCALE' if s.selected_name == 'Relative' else self.combo.get_active_id()
            val = s.selected_child.get_value(), mode
        self.destroy()
        return val


def white_borders(model, selection, pdfqueue):
    crop = []
    for path in selection:
        it = model.get_iter(path)
        p = model.get_value(it, 0)
        pdfdoc = pdfqueue[p.nfile - 1]

        page = pdfdoc.document.get_page(p.npage - 1)
        # Always render pages at 72 dpi whatever the zoom or scale of the page
        w, h = page.get_size()
        orig_crop = p.crop.rotated(-p.rotate_times(p.angle))

        first_col = int(w * orig_crop.left)
        last_col = min(int(w), int(w * (1 - orig_crop.right) + 1))
        first_row = int(h * orig_crop.top)
        last_row = min(int(h), int(h * (1 - orig_crop.bottom) + 1))
        w = int(w)
        h = int(h)
        thumbnail = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
        cr = cairo.Context(thumbnail)
        with pdfdoc.render_lock:
            page.render(cr)
        data = thumbnail.get_data().cast("i")
        whitecol = memoryview(b"\0" * (last_row - first_row) * 4).cast("i")
        whiterow = memoryview(b"\0" * (last_col - first_col) * 4).cast("i")
        crop_this_page = [0.0, 0.0, 0.0, 0.0]
        # Left
        for col in range(first_col, last_col):
            if data[col::w][first_row:last_row] != whitecol:
                crop_this_page[0] = (col - 1) / w
                break

        # Right
        for col in range(last_col - 1, first_col - 1, -1):
            if data[col::w][first_row:last_row] != whitecol:
                crop_this_page[1] = (w - col - 1) / w
                break

        # Top
        for row in range(first_row, last_row):
            if data[row * w : (row + 1) * w][first_col:last_col] != whiterow:
                crop_this_page[2] = (row - 1) / h
                break

        # Bottom
        for row in range(last_row - 1, first_row - 1, -1):
            if data[row * w : (row + 1) * w][first_col:last_col] != whiterow:
                crop_this_page[3] = (h - row - 1) / h
                break

        crop.append(Sides(*crop_this_page).rotated(p.rotate_times(p.angle)))
    return crop


class BlankPageDialog(BaseDialog):
    def __init__(self, size, window):
        super().__init__(title=_("Insert Blank Page"), parent=window)
        self.set_resizable(False)
        self.paper_widget = PaperSizeWidget(size)
        self.vbox.pack_start(self.paper_widget, True, True, 0)
        self.show_all()

    def run_get(self):
        result = self.run()
        r = None
        if result == Gtk.ResponseType.OK:
            r = self.paper_widget.get_value()
        self.destroy()
        return r


class MergePagesDialog(BaseDialog):

    def __init__(self, window, size, equal):
        super().__init__(title=_("Merge Pages"), parent=window)
        self.size = size
        self.set_resizable(False)
        max_margin = int(((14400 - max(*size)) / 2) * 25.4 / 72)
        self.marg = Gtk.SpinButton.new_with_range(0, max_margin, 1)
        self.marg.set_activates_default(True)
        self.marg.connect('value-changed', self.on_sb_value_changed)
        self.cols = Gtk.SpinButton.new_with_range(1, 2, 1)
        self.cols.set_activates_default(True)
        self.cols.connect('value-changed', self.on_sb_value_changed)
        self.rows = Gtk.SpinButton.new_with_range(1, 1, 1)
        self.rows.set_activates_default(True)
        self.rows.connect('value-changed', self.on_sb_value_changed)

        marg_lbl1 = Gtk.Label(_("Margin"), halign=Gtk.Align.START)
        marg_lbl2 = Gtk.Label(_("mm"), halign=Gtk.Align.START)
        cols_lbl1 = Gtk.Label(_("Columns"), halign=Gtk.Align.START)
        rows_lbl1 = Gtk.Label(_("Rows"), halign=Gtk.Align.START)
        grid1 = Gtk.Grid(column_spacing=12, row_spacing=12, margin=12, halign=Gtk.Align.CENTER)
        grid1.attach(marg_lbl1, 0, 1, 1, 1)
        grid1.attach(self.marg, 1, 1, 1, 1)
        grid1.attach(marg_lbl2, 2, 1, 1, 1)
        grid1.attach(cols_lbl1, 0, 2, 1, 1)
        grid1.attach(self.cols, 1, 2, 1, 1)
        grid1.attach(rows_lbl1, 0, 3, 1, 1)
        grid1.attach(self.rows, 1, 3, 1, 1)
        self.vbox.pack_start(grid1, False, False, 8)

        self.hor = Gtk.RadioButton(label=_("Horizontal"), group=None)
        vrt = Gtk.RadioButton(label=_("Vertical"), group=self.hor)
        self.l_r = Gtk.RadioButton(label=_("Left to Right"), group=None)
        r_l = Gtk.RadioButton(label=_("Right to Left"), group=self.l_r)
        self.t_b = Gtk.RadioButton(label=_("Top to Bottom"), group=None)
        b_t = Gtk.RadioButton(label=_("Bottom to Top"), group=self.t_b)
        grid2 = Gtk.Grid(column_spacing=6, row_spacing=12, margin=12, halign=Gtk.Align.CENTER)
        grid2.attach(self.hor, 0, 1, 1, 1)
        grid2.attach(vrt, 1, 1, 1, 1)
        grid2.attach(self.l_r, 0, 2, 1, 1)
        grid2.attach(r_l, 1, 2, 1, 1)
        grid2.attach(self.t_b, 0, 3, 1, 1)
        grid2.attach(b_t, 1, 3, 1, 1)
        frame1 = Gtk.Frame(label=_("Page Order"), margin=8)
        frame1.add(grid2)
        self.vbox.pack_start(frame1, False, False, 0)

        t = "" if equal else _("Non-uniform page size - using max size")
        warn_lbl = Gtk.Label(t, margin=8, wrap=True, width_chars=36, max_width_chars=36)
        self.vbox.pack_start(warn_lbl, False, False, 0)
        self.size_lbl = Gtk.Label(halign=Gtk.Align.CENTER, margin_bottom=16)
        self.vbox.pack_start(self.size_lbl, False, False, 0)
        self.show_all()

    def size_with_margin(self):
        width = self.size[0] + 2 * self.marg.get_value() * 72 / 25.4
        height = self.size[1] + 2 * self.marg.get_value() * 72 / 25.4
        return width, height

    def on_sb_value_changed(self, _button):
        width, height = self.size_with_margin()
        self.cols.set_range(1, 14400 // width)
        self.rows.set_range(1, 14400 // height)
        cols = int(self.cols.get_value())
        rows = int(self.rows.get_value())
        width = str(round(cols * width * 25.4 / 72, 1))
        height = str(round(rows * height * 25.4 / 72, 1))
        t =  _("Merged page size:") + " " + width + _("mm") + " \u00D7 " + height + _("mm")
        self.size_lbl.set_label(t)

    def run_get(self):
        self.cols.set_value(2)
        result = self.run()
        if result != Gtk.ResponseType.OK:
            self.destroy()
            return None
        cols = int(self.cols.get_value())
        rows = int(self.rows.get_value())
        range_cols = range(cols) if self.l_r.get_active() else range(cols)[::-1]
        range_rows = range(rows) if self.t_b.get_active() else range(rows)[::-1]
        if self.hor.get_active():
            order = [(row, col) for row in range_rows for col in range_cols]
        else:
            order = [(row, col) for col in range_cols for row in range_rows]
        self.destroy()
        return cols, rows, order, self.size_with_margin()


class _OffsetWidget(Gtk.Frame):
    def __init__(self, offset, dpage, lpage):
        super().__init__(shadow_type=Gtk.ShadowType.NONE)
        self.spinb_x = Gtk.SpinButton.new_with_range(0, 100, 1)
        self.spinb_x.set_activates_default(True)
        self.spinb_x.set_digits(1)
        self.spinb_x.connect('value-changed', self.spinb_val_changed)
        self.spinb_y = Gtk.SpinButton.new_with_range(0, 100, 1)
        self.spinb_y.set_activates_default(True)
        self.spinb_y.set_digits(1)
        self.spinb_y.connect('value-changed', self.spinb_val_changed)
        self.spinb_changed_callback = None

        lbl1_x = Gtk.Label(_("Horizontal offset"), halign=Gtk.Align.START)
        lbl2_x = Gtk.Label(_("%"), halign=Gtk.Align.START)
        lbl1_y = Gtk.Label(_("Vertical offset"), halign=Gtk.Align.START)
        lbl2_y = Gtk.Label(_("%"), halign=Gtk.Align.START)
        grid = Gtk.Grid(column_spacing=12, row_spacing=12, margin=12, halign=Gtk.Align.CENTER)
        grid.attach(lbl1_x, 0, 1, 1, 1)
        grid.attach(self.spinb_x, 1, 1, 1, 1)
        grid.attach(lbl2_x, 2, 1, 1, 1)
        grid.attach(lbl1_y, 0, 2, 1, 1)
        grid.attach(self.spinb_y, 1, 2, 1, 1)
        grid.attach(lbl2_y, 2, 2, 1, 1)
        self.add(grid)

        self.spinb_x.set_value(offset[0] * 100)
        self.spinb_y.set_value(offset[1] * 100)
        self.set_scale(dpage, lpage)

    def spinb_val_changed(self, spinbutton):
        if callable(self.spinb_changed_callback):
            self.spinb_changed_callback()

    def set_spinb_changed_callback(self, callback):
        self.spinb_changed_callback = callback

    def set_val(self, values):
        self.spinb_x.set_value(values.left * self.scale.left)
        self.spinb_y.set_value(values.top * self.scale.top)

    def get_val(self):
        """Get left, right, top, bottom offset from dest page edges."""
        xval = self.spinb_x.get_value()
        yval = self.spinb_y.get_value()
        return Sides(xval, 100 - xval, yval, 100 - yval) / self.scale

    def get_diff_offset(self):
        """Get the fraction of page size differenace at top-left."""
        return self.spinb_x.get_value() / 100, self.spinb_y.get_value() / 100

    def set_scale(self, dpage, lpage):
        """Set scale between 'destination page edge offset' and 'page size diff offset'."""
        dw, dh = dpage.size_in_pixel()
        lw, lh = lpage.size_in_pixel()
        scalex = 100 * dw / (dw - lw) if dw - lw != 0 else 1e10
        scaley = 100 * dh / (dh - lh) if dh - lh != 0 else 1e10
        self.scale = Sides(scalex, scalex, scaley, scaley)


class DrawingAreaWidget(Gtk.Box):
    """A widget which draws a page. It has tools for editing a rectangle (crop/hide/offset)."""

    def __init__(self, page, pdfqueue, spinbutton_widget=None, draw_on_page_func=None):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        page = page.duplicate()
        page.thumbnail = page.thumbnail if page.crop == Sides() else None
        page.resample = -1
        self.damodel = Gtk.ListStore(GObject.TYPE_PYOBJECT)
        self.damodel.append([page])
        self.pdfqueue = pdfqueue
        self.spinbutton_widget = spinbutton_widget
        self.padding = 25  # Around thumbnail
        self.click_pos = 0, 0
        self.click_val = [0] * 4
        self.x_po_rel_thmb = 0
        self.y_po_rel_thmb = 0
        self.x_po_rel_sw = 0
        self.y_po_rel_sw = 0
        self.cursor_name = 'default'
        self.rendering_thread = None
        self.render_id = None
        self.surface = None
        self.adjust_rect = [0] * 4
        self.allow_adjust_rect_resize = True
        self.handle_move_limits = True
        self.draw_on_page = draw_on_page_func
        self.da = Gtk.DrawingArea()
        self.da.set_events(self.da.get_events()
                              | Gdk.EventMask.BUTTON_PRESS_MASK
                              | Gdk.EventMask.BUTTON_RELEASE_MASK
                              | Gdk.EventMask.POINTER_MOTION_MASK)
        self.da.connect('draw', self.on_draw)
        self.da.connect('button-press-event', self.button_press_event)
        self.da.connect('button-release-event', self.button_release_event)
        self.da.connect('motion-notify-event', self.motion_notify_event)
        self.da.connect('size_allocate', self.size_allocate)

        self.sw = Gtk.ScrolledWindow()
        self.sw.set_size_request(self.padding + 100, self.padding + 100)
        self.sw.add(self.da)
        self.sw.connect('size_allocate', self.draw_page)
        self.sw.connect('scroll_event', self.sw_scroll_event)
        self.sw.connect('leave_notify_event', self.sw_leave_notify_event)
        self.pack_start(self.sw, True, True, 0)

        if self.spinbutton_widget is not None:
            self.spinbutton_widget.set_spinb_changed_callback(self.draw_page)
            self.pack_start(self.spinbutton_widget, False, False, 0)
            cb = Gtk.CheckButton(label=_("Show values"), margin=12, halign=Gtk.Align.CENTER)
            cb.connect('toggled', self.cb_show_val_toggled)
            cb.connect('realize', self.cb_realize)
            self.pack_start(cb, False, False, 0)

    def cb_realize(self, _cb):
        self.spinbutton_widget.hide()

    def cb_show_val_toggled(self, cb):
        self.spinbutton_widget.set_visible(cb.get_active())

    def store_pointer_location(self, sw, event):
        """Store pointer location relative to thumbnail and scrolled window."""
        ha = sw.get_hadjustment()
        thmb_x = event.x - self.padding + ha.get_value()
        self.x_po_rel_thmb = thmb_x / (ha.get_upper() - self.padding * 2)
        self.x_po_rel_sw = event.x / ha.get_page_size()

        va = sw.get_vadjustment()
        thmb_y = event.y - self.padding + va.get_value()
        self.y_po_rel_thmb = thmb_y / (va.get_upper() - self.padding * 2)
        self.y_po_rel_sw = event.y / va.get_page_size()

    def sw_scroll_event(self, sw, event):
        if not event.state & Gdk.ModifierType.CONTROL_MASK:
            return Gdk.EVENT_PROPAGATE
        w = max(p.width_in_pixel() for [p] in self.damodel)
        h = max(p.height_in_pixel() for [p] in self.damodel)
        maxfactor = (80000000 / (w * h)) ** .5  # Limit zoom at about 304Mb

        if event.direction == Gdk.ScrollDirection.SMOOTH:
            factor = round(min(1 - event.get_scroll_deltas()[2] * 0.3, maxfactor), 2)
        elif event.direction == Gdk.ScrollDirection.UP:
            factor = round(min(1.3, maxfactor), 2)
        elif event.direction == Gdk.ScrollDirection.DOWN:
            factor = 0.7
        else:
            return Gdk.EVENT_PROPAGATE
        self.store_pointer_location(sw, event)
        dpage = self.damodel[0][0]
        self.da.set_size_request((dpage.width_in_pixel() + self.padding * 2) * factor,
                                 (dpage.height_in_pixel() + self.padding * 2) * factor)
        return Gdk.EVENT_STOP

    def size_allocate(self, _da, da_rect):
        self.set_adjustment_values()
        self.set_zoom(da_rect)
        self.init_surface()
        self.silent_render()

    def set_adjustment_values(self):
        """Update adjustment values so it does zoom in at cursor."""
        ha = self.sw.get_hadjustment()
        thmb_x = (ha.get_upper() - self.padding * 2) * self.x_po_rel_thmb
        sw_x = ha.get_page_size() * self.x_po_rel_sw
        ha.set_value(self.padding + thmb_x - sw_x)

        va = self.sw.get_vadjustment()
        thmb_y = (va.get_upper() - self.padding * 2) * self.y_po_rel_thmb
        sw_y = va.get_page_size() * self.y_po_rel_sw
        va.set_value(self.padding + thmb_y - sw_y)

    def set_zoom(self, da_rect):
        dpage = self.damodel[0][0]
        thmb_max_w = da_rect.width - self.padding * 2
        thmb_max_h = da_rect.height - self.padding * 2
        zoom_x = thmb_max_w / dpage.width_in_points()
        zoom_y = thmb_max_h / dpage.height_in_points()
        for [page] in self.damodel:
            page.zoom = min(zoom_x, zoom_y)

    def silent_render(self):
        if self.render_id:
            GObject.source_remove(self.render_id)
        self.render_id = GObject.timeout_add(149, self.render)

    def quit_rendering(self):
        if self.rendering_thread is None:
            return False
        self.rendering_thread.quit = True
        self.rendering_thread.join(timeout=0.01)
        return self.rendering_thread.is_alive()

    def render(self):
        self.render_id = None
        alive = self.quit_rendering()
        if alive:
            self.silent_render()
            return
        self.rendering_thread = PDFRenderer(self.damodel, self.pdfqueue, [0, 1] , 1)
        self.rendering_thread.connect('update_thumbnail', self.update_thumbnail)
        self.rendering_thread.start()

    def update_thumbnail(self, _obj, ref, thumbnail, _zoom, _scale, _is_preview):
        if thumbnail is None:
            return
        path = ref.get_path()
        page = self.damodel[path][0]
        page.thumbnail = thumbnail
        self.draw_page()

    def button_press_event(self, _darea, event):
        self.click_pos = event.x, event.y
        if event.button == 2:
            self.set_cursor('move')
        elif event.button == 1 and self.spinbutton_widget is not None:
            self.click_val = self.spinbutton_widget.get_val()

    def button_release_event(self, _darea, event):
        sc = self.get_suggested_cursor(event)
        self.set_cursor(sc)

    def get_suggested_cursor(self, event):
        """Get appropriate cursor when moving mouse over adjust rect (crop/hide/offset)."""
        margin = 5
        r = self.adjust_rect
        if self.allow_adjust_rect_resize:
            w = r[0] - margin < event.x < r[0] + margin
            e = r[0] + r[2] - margin < event.x < r[0] + r[2] + margin
            n = r[1] - margin < event.y < r[1] + margin
            s = r[1] + r[3] - margin < event.y < r[1] + r[3] + margin
        else:
            w = e = n = s = False
        x_area = r[0] + margin < event.x < r[0] + r[2] - margin
        y_area = r[1] + margin < event.y < r[1] + r[3] - margin

        if n and w:
            cursor_name = 'nw-resize'
        elif s and w:
            cursor_name = 'sw-resize'
        elif s and e:
            cursor_name = 'se-resize'
        elif n and e:
            cursor_name = 'ne-resize'
        elif w and y_area:
            cursor_name = 'w-resize'
        elif e and y_area:
            cursor_name = 'e-resize'
        elif n and x_area:
            cursor_name = 'n-resize'
        elif s and x_area:
            cursor_name = 's-resize'
        elif x_area and y_area:
            cursor_name = 'move'
        else:
            cursor_name = 'default'
        return cursor_name

    def set_cursor(self, cursor_name):
        if cursor_name != self.cursor_name:
            self.cursor_name = cursor_name
            cursor = Gdk.Cursor.new_from_name(Gdk.Display.get_default(), cursor_name)
            self.get_window().set_cursor(cursor)

    def motion_notify_event(self, _darea, event):
        if event.state & Gdk.ModifierType.BUTTON2_MASK:
            self.pan_view(event)
            self.set_cursor('move')
        elif event.state & Gdk.ModifierType.BUTTON1_MASK:
            self.adjust_val(event)
        else:
            sc = self.get_suggested_cursor(event)
            self.set_cursor(sc)

    def pan_view(self, event):
        ha = self.sw.get_hadjustment()
        va = self.sw.get_vadjustment()
        ha.set_value(ha.get_value() + self.click_pos[0] - event.x)
        va.set_value(va.get_value() + self.click_pos[1] - event.y)

    def adjust_val(self, event):
        if self.spinbutton_widget is None:
            return
        left, right, top, bottom = self.spinbutton_widget.get_val()
        page = self.damodel[0][0]
        if self.cursor_name in ['w-resize', 'nw-resize', 'sw-resize', 'move']:
            left = self.click_val[0] + ((event.x - self.click_pos[0]) / page.width_in_pixel())
        if self.cursor_name in ['e-resize', 'ne-resize', 'se-resize', 'move']:
            right = self.click_val[1] - ((event.x - self.click_pos[0]) / page.width_in_pixel())
        if self.cursor_name in ['n-resize', 'nw-resize', 'ne-resize', 'move']:
            top = self.click_val[2] + ((event.y - self.click_pos[1]) / page.height_in_pixel())
        if self.cursor_name in ['s-resize', 'sw-resize', 'se-resize', 'move']:
            bottom = self.click_val[3] - ((event.y - self.click_pos[1]) / page.height_in_pixel())
        v = Sides(left, right, top, bottom)
        if self.cursor_name in ['move'] and self.handle_move_limits:
            v += Sides(*(min(0, v[i]) for i in [1, 0, 3, 2]))
        self.spinbutton_widget.set_spinb_changed_callback(None)
        self.spinbutton_widget.set_val(v)
        self.spinbutton_widget.set_spinb_changed_callback(self.draw_page)
        self.draw_page()

    def sw_leave_notify_event(self, _sw, event):
        if event.state & Gdk.ModifierType.BUTTON1_MASK:
            return
        self.set_cursor('default')

    def init_surface(self):
        aw = self.da.get_allocated_width()
        ah = self.da.get_allocated_height()
        self.surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, aw, ah)

    def on_draw(self, _darea, cr):
        if self.surface is not None:
            cr.set_source_surface(self.surface, 0, 0)
            cr.paint()

    def draw_page(self, _widget=None, _rect=None):
        """Draw the 'destination' thumbnail page."""
        if len(self.damodel) == 0 or self.surface is None:
            return
        dpage = self.damodel[0][0]
        if dpage.thumbnail is None:
            return
        cr = cairo.Context(self.surface)
        aw = self.da.get_allocated_width()
        ah = self.da.get_allocated_height()

        # Destination page rectangle
        dw = dpage.width_in_pixel()
        dh = dpage.height_in_pixel()
        dx = int(.5 + (aw - dw) / 2)
        dy = int(.5 + (ah - dh) / 2)

        # Page border
        cr.set_source_rgb(0, 0, 0)
        cr.rectangle(dx - 2, dy - 2, dw + 4, dh + 4)
        cr.fill_preserve()
        cr.clip()

        # Fill white paper
        cr.set_source_rgb(1, 1, 1)
        cr.rectangle(dx, dy, dw, dh)
        cr.fill()

        # Add the thumbnail
        (dw0, dh0) = (dh, dw) if dpage.angle in [90, 270] else (dw, dh)
        cr.translate(dx, dy)
        if dpage.angle > 0:
            cr.translate(dw / 2, dh / 2)
            cr.rotate(dpage.angle * pi / 180)
            cr.translate(-dw0 / 2, -dh0 / 2)
        tw, th = dpage.thumbnail.get_width(), dpage.thumbnail.get_height()
        tw, th = (th, tw) if dpage.angle in [90, 270] else (tw, th)
        cr.scale(dw / tw, dh / th)
        cr.set_source_surface(dpage.thumbnail)
        cr.get_source().set_filter(cairo.FILTER_FAST)
        cr.paint()

        cr.identity_matrix()
        cr.set_line_width(1)

        if dpage.hide != Sides():
            # Draw the hide rectangle. For the Hide dialog this will be the crop rectangle
            dwfull = dpage.scale * dpage.size.width * dpage.zoom
            dhfull = dpage.scale * dpage.size.height * dpage.zoom
            rx = round(dx + (dpage.hide.left - dpage.crop.left) * dwfull) - .5
            ry = round(dy + (dpage.hide.top - dpage.crop.top) * dhfull) - .5
            rw = round(dwfull * (1 - dpage.hide.left - dpage.hide.right)) + 1
            rh = round(dhfull * (1 - dpage.hide.top - dpage.hide.bottom)) + 1
            hide_rect = [rx, ry, rw, rh]

            cr.set_source_rgba(1, 1, 1, .7)
            cr.rectangle(*hide_rect)
            cr.stroke()
            cr.set_source_rgba(0, 0, 0, .3)
            cr.set_dash([8, 8])
            cr.rectangle(*hide_rect)
            cr.stroke()

            # Darken the area outside of rectangle
            cr.set_source_rgba(0, 0, 0, .3)
            cr.rectangle(dx, dy, dw, dh)
            cr.rectangle(*hide_rect)
            cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
            cr.fill()

        if callable(self.draw_on_page):
            self.adjust_rect = self.draw_on_page(cr, dx, dy, dw, dh, self.damodel)
            # Draw the adjust rectangle
            cr.identity_matrix()
            cr.set_source_rgb(1, 1, 1)
            cr.set_dash([])
            cr.rectangle(*self.adjust_rect)
            cr.stroke()
            cr.set_source_rgb(0, 0, 0)
            cr.set_dash([4.0, 4.0])
            cr.rectangle(*self.adjust_rect)
            cr.stroke()

        # Invalidate region
        ha = self.sw.get_hadjustment()
        va = self.sw.get_vadjustment()
        r = ha.get_value(), va.get_value(), ha.get_page_size(), va.get_page_size()
        self.da.queue_draw_area(*r)


class CropHideDialog():
    def __init__(self, window, selection, model, pdfqueue, is_unsaved, mode, update_val_func):
        title = _("Crop Margins") if mode == 'CROP' else _("Hide Margins")
        init_values = [getattr(model[row][0], mode.lower()) for row in selection]
        self.updated_values = init_values
        self.spinbutton_widget = _CropHideWidget(list(init_values[-1]), margin=8)
        page = model[selection[-1]][0]
        dawidget = DrawingAreaWidget(page, pdfqueue, self.spinbutton_widget, self.draw_on_page)
        page = dawidget.damodel[0][0]
        page.hide = page.crop if mode == 'HIDE' else page.hide
        page.crop = Sides()

        prepend_butt = (_("_Revert"), Gtk.ResponseType.REJECT, _("_Apply"), Gtk.ResponseType.APPLY)
        d = BaseDialog(title, window, prepend_butt)
        d.connect('response', self.on_response, selection, init_values, is_unsaved, update_val_func)
        d.vbox.pack_start(dawidget, True, True, 0)
        d.set_size_request(350, 500)
        d.show_all()

    def draw_on_page(self, cr, dx, dy, dw, dh, _damodel):
        """Draw on the thumbnail page."""
        v = self.spinbutton_widget.get_val()
        rx = round(dx + v.left * dw) - .5
        ry = round(dy + v.top * dh) - .5
        rw = round(dw * (1 - v.left - v.right)) + 1
        rh = round(dh * (1 - v.top - v.bottom)) + 1

        # Darken area outside of crop or hide rectangle
        cr.set_source_rgba(0, 0, 0, .3)
        cr.set_dash([])
        cr.rectangle(dx, dy, dw, dh)
        cr.rectangle(rx, ry, rw, rh)
        cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
        cr.fill()

        return [rx, ry, rw, rh]

    def on_response(self, dialog, response, selection, init_values, is_unsaved, update_val_func):
        if response == Gtk.ResponseType.REJECT:
            self.spinbutton_widget.set_val(init_values[-1])
            self.updated_values = init_values
            update_val_func(self.updated_values, selection, is_unsaved)
            return
        if response in [Gtk.ResponseType.OK, Gtk.ResponseType.APPLY]:
            new_val = self.spinbutton_widget.get_val()
            if any([new_val != val for val in self.updated_values]):
                self.updated_values = [new_val] * len(selection)
                update_val_func(self.updated_values, selection, True)
        if response == Gtk.ResponseType.APPLY:
            return
        dialog.destroy()


class PastePageLayerDialog():
    def __init__(self, window, dpage, lpage_stack, model, pdfqueue, mode, layer_pos):
        title = _("Overlay") if mode == 'OVERLAY' else _("Underlay")
        lpage = lpage_stack[0].duplicate()
        lpage.layerpages = [lp.duplicate() for lp in lpage_stack[1:]]
        lpage.zoom = dpage.zoom
        lpage.resample = -1
        lpage.thumbnail = None
        lpage.hide = Sides()
        self.spinbutton_widget = _OffsetWidget(layer_pos, dpage, lpage)
        dawidget = DrawingAreaWidget(dpage, pdfqueue, self.spinbutton_widget, self.draw_on_page)
        dawidget.allow_adjust_rect_resize = False
        dawidget.handle_move_limits = False
        dawidget.damodel.append([lpage])

        self.dialog = BaseDialog(title, window)
        self.dialog.vbox.pack_start(dawidget, True, True, 0)
        self.dialog.set_size_request(350, 500)
        self.dialog.show_all()

    def draw_on_page(self, cr, dx, dy, dw, dh, damodel):
        """Draw on the thumbnail page."""
        dpage, lpage = [page for [page] in damodel]
        if lpage.thumbnail is None:
            return [0] * 4
        self.spinbutton_widget.set_scale(dpage, lpage)

        # Layer page rectangle
        offset = self.spinbutton_widget.get_val()
        lx = round(dx + dw * offset.left)
        ly = round(dy + dh * offset.top)
        lw = round(lpage.zoom * lpage.width_in_points())
        lh = round(lpage.zoom * lpage.height_in_points())

        # Add the overlay/underlay
        (pw0, ph0) = (lh, lw) if lpage.angle in [90, 270] else (lw, lh)
        cr.translate(lx, ly)
        if lpage.angle > 0:
            cr.translate(lw / 2, lh / 2)
            cr.rotate(lpage.angle * pi / 180)
            cr.translate(-pw0 / 2, -ph0 / 2)
        ltw, lth = lpage.thumbnail.get_width(), lpage.thumbnail.get_height()
        ltw, lth = (lth, ltw) if lpage.angle in [90, 270] else (ltw, lth)
        cr.scale(lw / ltw, lh / lth)
        cr.set_source_surface(lpage.thumbnail)
        cr.get_source().set_filter(cairo.FILTER_FAST)
        cr.paint()

        return [lx - .5, ly - .5, lw + 1, lh + 1]

    def get_offset(self):
        """Get layer page x and y offset from top-left edge of the destination page.

        The offset is the fraction of space positioned at left and top of the pasted layer,
        where space is the difference in width and height between the layer and the page.
        """
        result = self.dialog.run()
        r = None
        if result == Gtk.ResponseType.OK:
            r = self.spinbutton_widget.get_diff_offset()
        self.dialog.destroy()
        return r

class RangeSelectDialog(BaseDialog):
    """ A dialog box to select a range of pages. """

    def __init__(self, window):
        super().__init__(title=_("Range Select"), parent=window)
        margin = 12
        range_frame = Gtk.Frame()
        self.set_resizable(False)
        grid = Gtk.Grid()
        grid.set_column_spacing(margin)
        grid.set_row_spacing(margin)
        grid.props.margin = margin
        range_frame.add(grid)
        label = Gtk.Label(label=_("Select range of pages: "))
        label.set_alignment(0.0, 0.5)
        grid.attach(label, 0, 0, 1, 1)
        self.range_entry_widget = Gtk.Entry()
        grid.attach(self.range_entry_widget, 1, 0, 1, 1)
        label = Gtk.Label(
            label=_(
                    'Use a comma to separate page numbers, '
                    'a dash to select a range of pages. \n'
                    'e.g. : "1,3,5-7,9"'
                )
        )
        label.props.margin = margin
        label.set_line_wrap(True)
        label.set_max_width_chars(38)
        grid.attach(label, 0, 1, 2, 1)
        self.vbox.pack_start(range_frame, False, False, 0)
        self.show_all()
        # Connect "changed" signal to function for checking user input
        self.range_entry_widget.connect('changed', self.on_changed)
        self.range_entry_widget.set_activates_default(True)

    def run_get(self):
        """ Open the dialog and return the selected range"""
        result = self.run()
        selected_range = None
        if result == Gtk.ResponseType.OK:
            selected_range = self.range_entry_widget.get_text()
        self.destroy()
        return selected_range

    def on_changed(self, *args):
        # Check each user entry to be one of the valid ones '0123456789,- '
        text = self.range_entry_widget.get_text()
        text = text.replace('--', '-').replace(',,', ',').replace('  ', ' ')
        self.range_entry_widget.set_text(''.join([char for char in text if char in '0123456789,- ']))



# Copyright (C) 2018-2019 Jerome Robert
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
undo/redo implemented with the memento design pattern.

The memento pattern is simpler than the command pattern.
Here the memory cost of memento is affordable because we
only store snapshots of the GtkListStore object, not of
the whole PDF files.
"""


class Manager(object):
    """
    Stack of GtkListStore models (Memento design pattern)
    """

    def __init__(self, app):
        self.app = app
        self.model = app.model
        self.states = []
        #: label of the previous undoable action
        self.label = None
        #: id of the current state
        self.current = 0
        self.undoaction = None
        self.redoaction = None

    def clear(self):
        self.states = []
        self.label = None
        self.current = 0

    def commit(self, label):
        """
        Must be called *BEFORE* each undoable actions
        :param label: label of the action
        """
        self.states = self.states[:self.current]
        self.states.append(([row[0].duplicate(False) for row in self.model], self.label,))
        self.current += 1
        self.label = label
        self.__refresh()

    def undo(self, _action, _param, _unused):
        if self.current == len(self.states):
            self.states.append(([row[0].duplicate(False) for row in self.model], self.label,))
        state, self.label = self.states[self.current - 1]
        self.__set_state(state)
        self.current -= 1
        self.app.set_unsaved(True)
        self.__refresh()

    def redo(self, _action, _param, _unused):
        state, self.label = self.states[self.current + 1]
        self.__set_state(state)
        self.current += 1
        self.app.set_unsaved(True)
        self.__refresh()

    def set_actions(self, undo, redo):
        self.undoaction = undo
        self.redoaction = redo
        self.__refresh()

    def __set_state(self, state):
        self.app.quit_rendering()
        self.app.iconview.unselect_all()
        with self.app.render_lock():
            self.model.clear()
            for page in state:
                # Do not reset the zoom level
                page.zoom = self.app.zoom_scale
                page.resample = -1
                self.model.append([page, page.description])
        self.app.update_iconview_geometry()
        self.app.update_max_zoom_level()
        self.app.retitle()
        self.app.update_statusbar()
        self.app.silent_render()

    def __refresh(self):
        if self.undoaction:
            self.undoaction.set_enabled(self.current >= 1)
        if self.redoaction:
            self.redoaction.set_enabled(self.current + 1 < len(self.states))
        # TODO: This is where to update the undo/redo menu items label to
        # show which action is going to be undone/redone. Because GtkImageMenuItem
        # will leads to many changes in translations this is currently postponed.



# Copyright (C) 2008-2017 Konstantinos Poulios, 2018-2019 Jerome Robert
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gi.repository import Gtk
from gi.repository import Gdk
from math import pi


class CellRendererImage(Gtk.CellRenderer):
    def __init__(self):
        Gtk.CellRenderer.__init__(self)
        self.th1 = 2.  # border thickness
        self.th2 = 3.  # shadow thickness
        self.page = None

    def set_page(self, page):
        self.page = page

    def get_geometry(self):
        rotation = int(self.page.angle) % 360
        rotation = round(rotation / 90) * 90
        w0 = self.page.thumbnail.get_width()
        h0 = self.page.thumbnail.get_height()
        if rotation == 90 or rotation == 270:
            w1, h1 = h0, w0
        else:
            w1, h1 = w0, h0
        w2 = self.page.width_in_pixel()
        h2 = self.page.height_in_pixel()

        return w0, h0, w1, h1, w2, h2, rotation

    def do_render(self, window, _widget, _background_area, cell_area, _expose_area):
        if not self.page.thumbnail:
            return

        w0, h0, w1, h1, w2, h2, rotation = self.get_geometry()
        th = int(2 * self.th1 + self.th2)
        w = w2 + th
        h = h2 + th

        x = cell_area.x
        y = cell_area.y
        if cell_area and w > 0 and h > 0:
            x += self.get_property('xalign') * (cell_area.width - w)
            y += self.get_property('yalign') * (cell_area.height - h)
        window.translate(int(0.5 + x), int(0.5 + y))

        # shadow
        window.set_source_rgb(0.5, 0.5, 0.5)
        window.rectangle(th, th, w2, h2)
        window.fill()

        # border
        window.set_source_rgb(0, 0, 0)
        window.rectangle(0, 0, w2 + 2 * self.th1, h2 + 2 * self.th1)
        window.fill()

        # image
        window.set_source_rgb(1, 1, 1)
        window.rectangle(self.th1, self.th1, w2, h2)
        window.fill_preserve()
        window.clip()

        window.translate(self.th1, self.th1)
        scale = self.page.resample * self.page.zoom
        window.scale(scale, scale)
        if rotation > 0:
            window.translate(w1 / 2, h1 / 2)
            window.rotate(rotation * pi / 180)
            window.translate(-w0 / 2, -h0 / 2)

        window.set_source_surface(self.page.thumbnail)
        window.paint()

    def do_get_size(self, _widget, cell_area=None):
        x = y = 0
        th = int(2 * self.th1 + self.th2)
        w = self.page.width_in_pixel() + th
        h = self.page.height_in_pixel() + th

        if cell_area and w > 0 and h > 0:
            x = self.get_property('xalign') * (
                    cell_area.width - w - self.get_property('xpad'))
            y = self.get_property('yalign') * (
                    cell_area.height - h - self.get_property('ypad'))
        w += 2 * self.get_property('xpad')
        h += 2 * self.get_property('ypad')
        return int(x), int(y), w, h


class IconviewCursor(object):
    """Move cursor, select pages and scroll with navigation keys."""
    def __init__(self, app):
        self.app = app
        self.model = self.app.iconview.get_model()
        self.move_cursor = True
        self.event = None
        self.iconview = None
        self.sel_start_page = None
        self.cursor_page_nr = 0
        self.cursor_page_nr_new = 0
        self.cursor_is_visible = False

    def handler(self, iconview, event):
        """Do all steps involved in cursor moving."""
        self.iconview = iconview
        self.event = event
        self.set_initial()
        self.set_selection_start_page()
        if self.move_cursor:
            self.move()
        self.select()
        self.scroll_iconview()

    def set_initial(self):
        """Cursor initial set."""
        self.move_cursor = True
        selection = self.iconview.get_selected_items()
        if len(selection) == 0:
            self.cursor_is_visible = False
        if selection and not self.iconview.get_cursor()[1] in selection:
            selection.sort(key=lambda x: x.get_indices()[0])
            self.iconview.set_cursor(selection[-1], None, False)
            if self.event.state & Gdk.ModifierType.SHIFT_MASK:
                self.sel_start_page = Gtk.TreePath.get_indices(selection[-1])[0]
                self.iconview.unselect_all()
        elif not self.iconview.get_cursor()[0]:
            if self.event.keyval == Gdk.KEY_End:
                self.cursor_page_nr_new = len(self.model) - 1
            else:
                self.cursor_page_nr_new = 0
            cursor_path_new = Gtk.TreePath.new_from_indices([self.cursor_page_nr_new])
            self.iconview.set_cursor(cursor_path_new, None, False)
            self.move_cursor = False
            self.iconview.unselect_all()

    def set_selection_start_page(self):
        """Set selection start page when shift + navigation keys are used."""
        if self.sel_start_page is None and self.event.state & Gdk.ModifierType.SHIFT_MASK:
            self.sel_start_page = Gtk.TreePath.get_indices(self.iconview.get_cursor()[1])[0]
            self.iconview.unselect_all()
        elif not self.event.state & Gdk.ModifierType.SHIFT_MASK:
            self.sel_start_page = None

    def move(self):
        """Move iconview cursor with navigation keys."""
        cursor_path = self.iconview.get_cursor()[1]
        self.cursor_page_nr = Gtk.TreePath.get_indices(cursor_path)[0]
        cursor_y = self.iconview.get_cell_rect(cursor_path)[1].y
        page_size = self.app.sw.get_vadjustment().get_page_size()
        spacing = self.iconview.get_row_spacing()
        columns_nr = self.iconview.get_columns()
        if self.event.keyval == Gdk.KEY_Up:
            step = 0 if self.cursor_page_nr - columns_nr < 0 else columns_nr
            self.cursor_page_nr_new = self.cursor_page_nr - step
        elif self.event.keyval == Gdk.KEY_Down:
            step = 0 if self.cursor_page_nr + columns_nr > len(self.model) - 1 else columns_nr
            self.cursor_page_nr_new = self.cursor_page_nr + step
        elif self.event.keyval == Gdk.KEY_Left:
            if self.iconview.get_direction() == Gtk.TextDirection.LTR:
                self.cursor_page_nr_new = max(self.cursor_page_nr - 1, 0)
            else:
                self.cursor_page_nr_new = min(self.cursor_page_nr + 1, len(self.model) - 1)
        elif self.event.keyval == Gdk.KEY_Right:
            if self.iconview.get_direction() == Gtk.TextDirection.LTR:
                self.cursor_page_nr_new = min(self.cursor_page_nr + 1, len(self.model) - 1)
            else:
                self.cursor_page_nr_new = max(self.cursor_page_nr - 1, 0)
        elif self.event.keyval == Gdk.KEY_Home:
            self.cursor_page_nr_new = 0
        elif self.event.keyval == Gdk.KEY_End:
            self.cursor_page_nr_new = len(self.model) - 1
        elif self.event.keyval in [Gdk.KEY_Page_Up, Gdk.KEY_KP_Page_Up]:
            for self.cursor_page_nr_new in range(self.cursor_page_nr, -1, -columns_nr):
                new_path = Gtk.TreePath.new_from_indices([self.cursor_page_nr_new])
                new_y = self.iconview.get_cell_rect(new_path)[1].y
                if new_y < cursor_y - page_size - spacing:
                    self.cursor_page_nr_new += columns_nr
                    break
        elif self.event.keyval in [Gdk.KEY_Page_Down, Gdk.KEY_KP_Page_Down]:
            for self.cursor_page_nr_new in range(self.cursor_page_nr, len(self.model), columns_nr):
                new_path = Gtk.TreePath.new_from_indices([self.cursor_page_nr_new])
                new_y = self.iconview.get_cell_rect(new_path)[1].y
                if new_y > cursor_y + page_size + spacing:
                    self.cursor_page_nr_new -= columns_nr
                    break
        self.iconview.unselect_path(cursor_path)
        if len(self.model) > 1 and not self.cursor_is_visible:
            step = 1 if self.iconview.get_item_column(cursor_path) == 0 else -1
            self.iconview.emit('move-cursor', Gtk.MovementStep.VISUAL_POSITIONS, step)
            self.cursor_is_visible = True
            self.iconview.unselect_all()
        cursor_path_new = Gtk.TreePath.new_from_indices([self.cursor_page_nr_new])
        sw_hpos = self.app.sw.get_hadjustment().get_value()
        self.iconview.set_cursor(cursor_path_new, None, False)
        self.app.sw.get_hadjustment().set_value(sw_hpos)

    def select(self):
        """Select iconview pages with shift + navigation keys."""
        if self.event.state & Gdk.ModifierType.SHIFT_MASK:
            if self.cursor_page_nr_new > self.sel_start_page > self.cursor_page_nr:
                for page_nr in range(self.cursor_page_nr, self.sel_start_page):
                    path = Gtk.TreePath.new_from_indices([page_nr])
                    self.iconview.unselect_path(path)
            elif self.cursor_page_nr_new < self.sel_start_page < self.cursor_page_nr:
                for page_nr in range(self.sel_start_page, self.cursor_page_nr):
                    path = Gtk.TreePath.new_from_indices([page_nr])
                    self.iconview.unselect_path(path)
            elif self.cursor_page_nr_new == self.sel_start_page:
                self.iconview.unselect_all()
            step = -1 if self.cursor_page_nr_new < self.sel_start_page else 1
            for page_nr in range(self.cursor_page_nr_new, self.cursor_page_nr + step, step):
                path = Gtk.TreePath.new_from_indices([page_nr])
                self.iconview.unselect_path(path)
            for page_nr in range(self.sel_start_page, self.cursor_page_nr_new + step, step):
                path = Gtk.TreePath.new_from_indices([page_nr])
                self.iconview.select_path(path)
        else:
            self.iconview.unselect_all()
            cursor_path_new = Gtk.TreePath.new_from_indices([self.cursor_page_nr_new])
            self.iconview.select_path(cursor_path_new)

    def scroll_iconview(self):
        """Scroll in order to keep cursor visible in window."""
        sw_vadj = self.app.sw.get_vadjustment()
        sw_vpos = sw_vadj.get_value()
        cursor_path_new = Gtk.TreePath.new_from_indices([self.cursor_page_nr_new])
        cell_rect = self.iconview.get_cell_rect(cursor_path_new)[1]
        cell_height = cell_rect.height
        cell_y = self.iconview.convert_widget_to_bin_window_coords(cell_rect.x, cell_rect.y)[1]
        sw_height = self.app.sw.get_allocated_height()
        columns_nr = self.iconview.get_columns()
        up = self.cursor_page_nr_new < self.cursor_page_nr
        at_min = self.cursor_page_nr_new - columns_nr < 0
        down = self.cursor_page_nr_new > self.cursor_page_nr
        at_max = self.cursor_page_nr_new + columns_nr > len(self.model) - 1
        if up or at_min:
            sw_vpos = min(max(sw_vpos, cell_y + cell_height - sw_height), cell_y)
        elif down or at_max:
            sw_vpos = max(min(sw_vpos, cell_y), cell_y + cell_height - sw_height)
        sw_vadj.set_value(sw_vpos)


class IconviewDragSelect:
    """
    Drag-select when clicking between items and dragging mouse pointer.

    Click + drag-select selects the range from click location to drag location.
    Shift + click + drag-select adds more items to selection.
    Control + click + drag-select toggles selection.
    """
    def __init__(self, app):
        self.app = app
        self.iconview = app.iconview
        self.model = app.iconview.get_model()
        self.range_start = 0
        self.range_end = 0
        self.click_location = None
        self.cursor_name_old = 'default'

    def click(self, event):
        """Store the click location."""
        if len(self.model) == 0:
            self.click_location = None
            return
        self.click_location = self.get_location(event.x, event.y)
        if self.click_location:
            self.set_mouse_cursor('text')
            self.range_start = int(self.click_location + 0.5)
            self.range_end = self.range_start
            self.control_is_pressed = event.state & Gdk.ModifierType.CONTROL_MASK
            self.shift_is_pressed = event.state & Gdk.ModifierType.SHIFT_MASK
            if self.control_is_pressed or self.shift_is_pressed:
                self.selection_list = []
                for row in self.model:
                    self.selection_list.append(self.iconview.path_is_selected(row.path))

    def motion(self, event=None, step=0):
        """Get drag location and select or deselect items."""
        if not self.click_location:
            return False
        sw_vadj = self.app.sw.get_vadjustment()
        sw_vpos = sw_vadj.get_value()
        if event:
            self.event_x = event.x
            event_y = self.event_y = event.y + step
            self.sw_vpos = sw_vpos
            if not event.state & Gdk.ModifierType.CONTROL_MASK:
                self.control_is_pressed = False
            if not event.state & Gdk.ModifierType.SHIFT_MASK:
                self.shift_is_pressed = False
        else:
            event_y = self.event_y - self.sw_vpos + sw_vpos + step
        drag_location = self.get_location(self.event_x, event_y)
        if drag_location is None:
            return False
        selection_changed = self.select(drag_location)
        return selection_changed

    def select(self, drag_location):
        """Select or deselect items between click location and current mouse pointer location."""
        range_start_old = self.range_start
        range_end_old = self.range_end
        if drag_location > self.click_location:
            self.range_start = int(self.click_location + 0.5)
            self.range_end = int(drag_location + 1)
        else:
            self.range_start = int(drag_location + 0.5)
            self.range_end = int(self.click_location + 1)
        if self.range_start == range_start_old and self.range_end == range_end_old:
            return False
        changed_range_start = min(self.range_start, range_start_old)
        changed_range_end = max(self.range_end, range_end_old)
        for page_nr in range(changed_range_start, changed_range_end):
            path = Gtk.TreePath.new_from_indices([page_nr])
            if self.control_is_pressed:
                if page_nr in range(self.range_start, self.range_end):
                    if self.selection_list[page_nr]:
                        self.iconview.unselect_path(path)
                    else:
                        self.iconview.select_path(path)
                else:
                    if self.selection_list[page_nr]:
                        self.iconview.select_path(path)
                    else:
                        self.iconview.unselect_path(path)
            elif self.shift_is_pressed:
                if page_nr in range(self.range_start, self.range_end):
                    self.iconview.select_path(path)
                elif not self.selection_list[page_nr]:
                    self.iconview.unselect_path(path)
            else:
                if page_nr in range(self.range_start, self.range_end):
                    self.iconview.select_path(path)
                else:
                    self.iconview.unselect_path(path)
        return True

    def get_location(self, x, y):
        """
        Get mouse pointer location.

        E.g. Location is 2.0 when pointer is on item 2.
             Location is 2.5 when pointer is between item 2 and 3.
        """
        last = self.iconview.get_cell_rect(self.model[-1].path)[1]
        if last.width < 50:
            # Avoid seg fault if calling get_path_at_pos before iconview is ready
            return None
        last_x, last_y = self.iconview.convert_widget_to_bin_window_coords(last.x, last.y)
        x_step = last.width - 2 * self.iconview.get_item_padding() - 1
        y_step = self.iconview.get_row_spacing() + 2 * self.iconview.get_item_padding() + 1
        search_pos = [('XY', x, y),
                      ('Right', x + x_step, y),
                      ('Left', x - x_step, y),
                      ('Below', x, y + y_step),
                      ('Below', x + x_step, y + y_step),
                      ('Below', x - x_step, y + y_step),
                      ('Above', x, y - y_step),
                      ('Above', x + x_step, y - y_step),
                      ('Above', x - x_step, y - y_step),
                      ('Zero', 0, 0)]
        ind = 0
        for pos, x_s, y_s in search_pos:
            path = self.iconview.get_path_at_pos(x_s, y_s)
            if path:
                ind = Gtk.TreePath.get_indices(path)[0]
                break
        add = 0.5 if self.iconview.get_direction() == Gtk.TextDirection.LTR else -0.5
        if pos == 'XY':
            location = ind
        elif pos == 'Right':
            location = ind - add
        elif pos == 'Left':
            location = ind + add
        elif pos == 'Below':
            location = ind - self.iconview.get_item_column(path) - add
        elif pos == 'Above':
            location = ind + self.iconview.get_columns() - self.iconview.get_item_column(path) - add
        elif (y > last_y + last.height) or (y > last_y and x > last_x + last.width):
            location = len(self.model) - add
        elif y < 0:
            location = -add
        else:
            return None
        return min(location, len(self.model) - add)

    def set_mouse_cursor(self, cursor_name):
        """Set the cursor type specified by cursor_name."""
        if cursor_name == self.cursor_name_old:
            return
        cursor = Gdk.Cursor.new_from_name(Gdk.Display.get_default(), cursor_name)
        self.iconview.get_window().set_cursor(cursor)
        self.cursor_name_old = cursor_name

    def end(self):
        self.set_mouse_cursor('default')
        self.click_location = None


class IconviewPanView:
    """Pan the view when pressing mouse wheel and moving mouse."""
    def __init__(self, app):
        self.iconview = app.iconview
        self.sw_hadj = app.sw.get_hadjustment()
        self.sw_vadj = app.sw.get_vadjustment()
        self.cursor_name = 'default'

    def click(self, event):
        self.cursor_name = 'move'
        cursor = Gdk.Cursor.new_from_name(Gdk.Display.get_default(), self.cursor_name)
        self.iconview.get_window().set_cursor(cursor)
        self.click_x = event.x
        self.click_y = event.y

    def motion(self, event):
        if self.cursor_name == 'default':
            return
        self.sw_hadj.set_value(self.sw_hadj.get_value() + self.click_x - event.x)
        self.sw_vadj.set_value(self.sw_vadj.get_value() + self.click_y - event.y)

    def end(self):
        if self.cursor_name == 'default':
            return
        self.cursor_name = 'default'
        cursor = Gdk.Cursor.new_from_name(Gdk.Display.get_default(), self.cursor_name)
        self.iconview.get_window().set_cursor(cursor)
